from unittest import mock
from unittest.mock import MagicMock

import pytest

import dataquality
import dataquality.core.log
from dataquality.schemas.task_type import TaskType


def test_finish_no_init() -> None:
    """
    Tests finish without an init call
    """
    dataquality.config.current_run_id = dataquality.config.current_project_id = None
    with pytest.raises(AssertionError):
        dataquality.finish()


@mock.patch.object(dataquality.core.init.ApiClient, "wait_for_run")
def test_wait_for_run(mock_client: MagicMock) -> None:
    """
    Tests that wait_for_run calls ApiClient
    """
    dataquality.wait_for_run(project_name="Carrots", run_name="Rhubarb")
    mock_client.assert_called_once_with(project_name="Carrots", run_name="Rhubarb")


@mock.patch.object(
    dataquality.core.init.ApiClient,
    "get_run_status",
    return_value={"status": "started"},
)
def test_get_run_status(mock_client: MagicMock) -> None:
    """
    Tests that get_run_status calls ApiClient
    """
    status = dataquality.get_run_status(project_name="Carrots", run_name="Rhubarb")
    mock_client.assert_called_once_with(project_name="Carrots", run_name="Rhubarb")
    assert status.get("status") == "started"


@mock.patch.object(dataquality.core.finish, "_version_check")
@mock.patch.object(dataquality.core.finish, "_reset_run")
@mock.patch.object(dataquality.core.finish, "upload_dq_log_file")
@mock.patch.object(dataquality.clients.api.ApiClient, "make_request")
@mock.patch.object(
    dataquality.core.finish.dataquality,
    "get_data_logger",
)
@mock.patch.object(dataquality.core.finish, "wait_for_run")
def test_finish_waits_default(
    mock_wait_for_run: MagicMock,
    mock_get_data_logger: MagicMock,
    mock_make_request: MagicMock,
    mock_upload_log_file: MagicMock,
    mock_reset_run: MagicMock,
    mock_version_check: MagicMock,
    set_test_config,
) -> None:
    set_test_config(task_type=TaskType.text_classification)
    mock_get_data_logger.return_value = MagicMock()
    dataquality.finish()
    mock_wait_for_run.assert_called_once()


@mock.patch.object(dataquality.core.finish, "_version_check")
@mock.patch.object(dataquality.core.finish, "_reset_run")
@mock.patch.object(dataquality.core.finish, "upload_dq_log_file")
@mock.patch.object(dataquality.clients.api.ApiClient, "make_request")
@mock.patch.object(
    dataquality.core.finish.dataquality,
    "get_data_logger",
)
@mock.patch.object(dataquality.core.finish, "wait_for_run")
def test_finish_no_waits_when_false(
    mock_wait_for_run: MagicMock,
    mock_get_data_logger: MagicMock,
    mock_make_request: MagicMock,
    mock_upload_log_file: MagicMock,
    mock_reset_run: MagicMock,
    mock_version_check: MagicMock,
    set_test_config,
) -> None:
    set_test_config(task_type=TaskType.text_classification)
    mock_get_data_logger.return_value = MagicMock()
    dataquality.finish(wait=False)
    mock_wait_for_run.assert_not_called()



import os
from typing import Any, Dict, Optional

from pydantic import UUID4

import dataquality
from dataquality.clients.api import ApiClient
from dataquality.core._config import config
from dataquality.schemas import RequestType, Route
from dataquality.schemas.job import JobName
from dataquality.schemas.task_type import TaskType
from dataquality.utils.dq_logger import DQ_LOG_FILE_HOME, upload_dq_log_file
from dataquality.utils.helpers import check_noop
from dataquality.utils.thread_pool import ThreadPoolManager
from dataquality.utils.version import _version_check

api_client = ApiClient()


@check_noop
def finish(
    last_epoch: Optional[int] = None, wait: bool = True
) -> Optional[Dict[str, Any]]:
    """
    Finishes the current run and invokes a job

    :param last_epoch: If set, only epochs up to this value will be uploaded/processed
        This is inclusive, so setting last_epoch to 5 would upload epochs 0,1,2,3,4,5
    :param wait: If true, after uploading the data, this will wait for the
        run to be processed by the Galileo server. If false, you can manually wait
        for the run by calling `dq.wait_for_run()` Default True
    """
    ThreadPoolManager.wait_for_threads()
    assert config.current_project_id, "You must have an active project to call finish"
    assert config.current_run_id, "You must have an active run to call finish"
    data_logger = dataquality.get_data_logger()
    data_logger.validate_labels()

    _version_check()

    if data_logger.non_inference_logged():
        _reset_run(config.current_project_id, config.current_run_id, config.task_type)

    data_logger.upload(last_epoch)
    data_logger._cleanup()
    upload_dq_log_file()

    body = dict(
        project_id=str(config.current_project_id),
        run_id=str(config.current_run_id),
        labels=data_logger.logger_config.labels,
        tasks=data_logger.logger_config.tasks,
    )
    if data_logger.logger_config.inference_logged:
        body.update(
            job_name=JobName.inference,
            non_inference_logged=data_logger.non_inference_logged(),
        )
    res = api_client.make_request(
        RequestType.POST, url=f"{config.api_url}/{Route.jobs}", body=body
    )
    print(
        f"Job {res['job_name']} successfully submitted. Results will be available "
        f"soon at {res['link']}"
    )
    # Reset all config variables
    data_logger.logger_config.reset()
    if wait:
        wait_for_run()
    return res


@check_noop
def wait_for_run(
    project_name: Optional[str] = None, run_name: Optional[str] = None
) -> None:
    """
    Waits until a specific project run transitions from started to finished.
    Defaults to the current run if project_name and run_name are empty.
    Raises error if only one of project_name and run_name is passed in.

    :param project_name: The project name. Default to current project if not passed in.
    :param run_name: The run name. Default to current run if not passed in.
    :return: None. Function returns after the run transitions to `finished`
    """
    return api_client.wait_for_run(project_name=project_name, run_name=run_name)


@check_noop
def get_run_status(
    project_name: Optional[str] = None, run_name: Optional[str] = None
) -> Dict[str, Any]:
    """
    Returns the status of a specified project run.
    Defaults to the current run if project_name and run_name are empty.
    Raises error if only one of project_name and run_name is passed in.

    :param project_name: The project name. Default to current project if not passed in.
    :param run_name: The run name. Default to current run if not passed in.
    :return: Dict[str, Any]. Response will have key `status` with value corresponding
      to the status of the run. Other info, such as `timestamp`, may be included.
    """
    return api_client.get_run_status(project_name=project_name, run_name=run_name)


@check_noop
def _reset_run(
    project_id: UUID4, run_id: UUID4, task_type: Optional[TaskType] = None
) -> None:
    """Clear the data in minio before uploading new data

    If this is a run that already existed, we want to fully overwrite the old data.
    We can do this by deleting the run and recreating it with the same name, which will
    give it a new ID
    """
    old_run_id = run_id
    api_client.reset_run(project_id, old_run_id, task_type)
    project_dir = (
        f"{dataquality.get_data_logger().LOG_FILE_DIR}/{config.current_project_id}"
    )
    # All of the logged user data is to the old run ID, so rename it to the new ID
    os.rename(f"{project_dir}/{old_run_id}", f"{project_dir}/{config.current_run_id}")
    # Move std logs as well
    os.rename(
        f"{DQ_LOG_FILE_HOME}/{old_run_id}",
        f"{DQ_LOG_FILE_HOME}/{config.current_run_id}",
    )
    config.update_file_config()



from collections.abc import Callable
from enum import Enum

import requests


class RequestType(str, Enum):
    GET = "get"
    POST = "post"
    PUT = "put"
    DELETE = "delete"

    @staticmethod
    def get_method(request: str) -> Callable:
        return getattr(requests, request)



from typing import Optional, Tuple, Type

import h5py
import numpy as np

HDF5_STORE = "hdf5_store.hdf5"


class HDF5Store(object):
    """
    Simple class to append value to a hdf5 file on disc.

    Used to concatenate HDF5 files for vaex

    Params:
        datapath: filepath of h5 file
        dataset: dataset name within the file
        shape: dataset shape (not counting main/batch axis)
        dtype: numpy dtype

    Usage:
        hdf5_store = HDF5Store('/tmp/hdf5_store.h5','X', shape=(20,20,3))
        x = np.random.random(hdf5_store.shape)
        hdf5_store.append(x)
        hdf5_store.append(x)

    Adapted From https://gist.github.com/wassname/a0a75f133831eed1113d052c67cf8633
    """

    def __init__(
        self,
        datapath: str,
        dataset: str,
        shape: Tuple,
        dtype: Type = np.float32,
        compression: Optional[str] = None,
        chunk_len: int = 1,
    ):
        self.datapath = datapath
        self.dataset = dataset
        self.shape = shape
        self.i = 0

        with h5py.File(self.datapath, mode="a") as h5f:
            self.dset = h5f.create_dataset(
                dataset,
                shape=(0,) + shape,
                maxshape=(None,) + shape,
                dtype=dtype,
                compression=compression,
                chunks=(chunk_len,) + shape,
            )

    def append(self, values: np.ndarray) -> None:
        with h5py.File(self.datapath, mode="a") as h5f:
            dset = h5f[self.dataset]
            dset.resize((self.i + values.shape[0],) + self.shape)
            dset[self.i :] = values
            self.i += values.shape[0]
            h5f.flush()



import itertools
from collections import defaultdict
from enum import Enum, unique
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

import pandas as pd
import pyarrow as pa
import vaex
from vaex.dataframe import DataFrame

from dataquality.exceptions import GalileoException
from dataquality.loggers.data_logger.base_data_logger import (
    ITER_CHUNK_SIZE,
    BaseGalileoDataLogger,
    DataSet,
    MetasType,
    MetaType,
)
from dataquality.loggers.logger_config.text_ner import text_ner_logger_config
from dataquality.schemas import __data_schema_version__
from dataquality.schemas.dataframe import BaseLoggerInOutFrames
from dataquality.schemas.ner import NERColumns as NERCols
from dataquality.schemas.ner import TaggingSchema
from dataquality.schemas.split import Split
from dataquality.utils.vaex import rename_df


@unique
class GalileoDataLoggerAttributes(str, Enum):
    texts = "texts"
    text_token_indices = "text_token_indices"
    text_token_indices_flat = "text_token_indices_flat"
    gold_spans = "gold_spans"
    ids = "ids"
    # mixin restriction on str (due to "str".split(...))
    split = "split"  # type: ignore
    meta = "meta"  # Metadata columns for logging

    @staticmethod
    def get_valid() -> List[str]:
        return list(map(lambda x: x.value, GalileoDataLoggerAttributes))


class TextNERDataLogger(BaseGalileoDataLogger):
    """
    Class for logging input data/metadata of Text NER models to Galileo.

    * text: The raw text inputs for model training. List[str]

    * text_token_indices: Token boundaries of text. List[List[Tuple(int, int)]].
    Used to convert the gold_spans into token level spans internally. For each sample,
    the boundary of a token will contain the start and end character index of word in
    the `text` to which the said token belongs.

    * gold_spans: Gold spans for the text at character level indices.
    The list of spans in a sample with their start and end indexes, and the label.
    Indexes start at 0 and are [inclusive, exclusive) for [start, end) respectively.
    List[List[dict]].

    * ids: Optional unique indexes for each record. If not provided, will default to
    the index of the record. Optional[List[int]]

    * split: The split of training/test/validation

    * meta: Dict[str, List]. Any metadata information you want to log at a per sample
    (text input) level. This could be a string (len <= 50), a float or an int.
    Each sample can have up to 50 meta fields.

    # number of samples in the list must be the same length as the number of text
    # samples logged
    Format {"sample_importance": [0.2, 0.5, 0.99, ...]}

    ex:
    .. code-block:: python

        labels = ["B-PER", "I-PER", "B-LOC", "I-LOC", "O"]
        dq.set_labels_for_run(labels = labels)

        # One of (IOB2, BIO, IOB, BILOU, BILOES)
        dq.set_tagging_schema(tagging_schema: str = "BIO")

        texts: List[str] = [
            "The president is Joe Biden",
            "Joe Biden addressed the United States on Monday"
        ]

        gold_spans: List[List[dict]] = [
            [
                {"start":17, "end":27, "label":"person"}  # "Joe Biden"
            ],
            [
                {"start":0, "end":10, "label":"person"},    # "Joe Biden"
                {"start":30, "end":41, "label":"location"}  # "United States"
            ]
        ]

        text_token_indices: [[(0, 3), (4, 13), (14, 16), (17, 20), (21, 27), (21, 27)],
                [...]]
        ids: List[int] = [0, 1]
        meta = {"sample_quality": [5.3, 1.1]}
        split = "training"

        dq.log_data_samples(
            texts=texts, text_token_indices=text_token_indices,
            gold_spans=gold_spans, meta=meta, ids=ids, split=split
        )
    """

    DATA_FOLDER_EXTENSION = {"emb": "hdf5", "prob": "hdf5", "data": "arrow"}

    __logger_name__ = "text_ner"
    logger_config = text_ner_logger_config

    def __init__(
        self,
        texts: List[str] = None,
        text_token_indices: List[List[Tuple[int, int]]] = None,
        gold_spans: List[List[Dict]] = None,
        ids: List[int] = None,
        split: str = None,
        meta: MetasType = None,
    ) -> None:
        """Create data logger.

        :param texts: The raw text inputs for model training. List[str]
        :param text_token_indices: Token boundaries of text. List[Tuple(int, int)].
        Used to convert the gold_spans into token level spans internally.
        t[0] indicates the start index of the span and t[1] is the end index (exclusive)
        :param gold_spans: The model-level gold spans over the char index of `text`
        :param ids: Optional unique indexes for each record. If not provided, will
        default to the index of the record. Optional[List[Union[int,str]]]
        :param split: The split for training/test/validation
        """
        super().__init__(meta)
        # Need to compare to None because they may be np arrays which cannot be
        # evaluated with bool directly
        self.texts = texts if texts is not None else []
        self.text_token_indices = (
            text_token_indices if text_token_indices is not None else []
        )
        self.gold_spans = gold_spans if gold_spans is not None else []
        self.ids = ids if ids is not None else []
        self.split = split
        self.text_token_indices_flat: List[List[int]] = []

    @staticmethod
    def get_valid_attributes() -> List[str]:
        """
        Returns a list of valid attributes that GalileoModelConfig accepts
        :return: List[str]
        """
        return GalileoDataLoggerAttributes.get_valid()

    def log_data_samples(
        self,
        *,
        texts: List[str],
        ids: List[int],
        text_token_indices: List[List[Tuple[int, int]]] = None,
        gold_spans: List[List[Dict]] = None,
        split: Optional[Split] = None,
        meta: Optional[MetasType] = None,
        **kwargs: Any,  # For typing
    ) -> None:
        """Log input samples for text classification

        :param texts: List[str] text samples
        :param ids: List[int,str] IDs for each text sample
        :param text_token_indices: List[List[Tuple(int, int)]]. Token boundaries of each
            text sample, 1 list per sample.
            Used to convert the gold_spans into token level spans internally.
            t[0] indicates the start index of the span and t[1] is the end index
            (exclusive). Required if split is not inference
        :param gold_spans: List[List[Dict]] The model-level gold spans over the char
            index for each text sample. 1 List[Dict] per text sample.
            "start", "end", "label" are the required keys
            Required if split is not inference
        :param ids: Optional unique indexes for each record. If not provided, will
        :param split: train/test/validation/inference. Can be set here or via
            dq.set_split
        :param meta: Dict[str, List[str, int, float]]. Metadata for each text sample
            Format is the {"metadata_field_name": [metdata value per sample]}
        """
        self.validate_kwargs(kwargs)
        self.texts = texts
        self.ids = ids
        self.split = split
        self.text_token_indices = (
            text_token_indices if text_token_indices is not None else []
        )
        self.gold_spans = gold_spans if gold_spans is not None else []
        self.meta = meta or {}
        self.log()

    def log_data_sample(
        self,
        *,
        text: str,
        id: int,
        text_token_indices: List[Tuple[int, int]] = None,
        gold_spans: List[Dict] = None,
        split: Optional[Split] = None,
        meta: Optional[MetaType] = None,
        **kwargs: Any,
    ) -> None:
        """
        Log a single input sample for text classification

        :param text: str the text sample
        :param id: The sample ID
        :param text_token_indices: List[Tuple(int, int)]. Token boundaries of the
            text sample. Used to convert gold_spans into token level spans internally.
            t[0] indicates the start index of the span and t[1] is the end index
            (exclusive). Required if split is not inference
        :param gold_spans: List[Dict] The model-level gold spans over the char
            index of the text sample. "start", "end", "label" are the required keys
            Required if split is not inference
        :param split: train/test/validation/inference. Can be set here or via
            dq.set_split
        :param meta: Dict[str, Union[str, int, float]]. Metadata for the text sample
            Format is the {"metadata_field_name": metadata_field_value}
        """
        self.validate_kwargs(kwargs)
        self.texts = [text]
        self.ids = [id]
        self.split = split
        self.text_token_indices = (
            [text_token_indices] if text_token_indices is not None else []
        )
        self.gold_spans = [gold_spans] if gold_spans is not None else []
        self.meta = {i: [meta[i]] for i in meta} if meta else {}
        self.log()

    def log_dataset(
        self,
        dataset: DataSet,
        *,
        batch_size: int = ITER_CHUNK_SIZE,
        text: Union[str, int] = "text",
        id: Union[str, int] = "id",
        text_token_indices: Union[str, int] = "text_token_indices",
        gold_spans: Union[str, int] = "gold_spans",
        split: Optional[Split] = None,
        meta: Optional[List[Union[str, int]]] = None,
        **kwargs: Any,
    ) -> None:
        """
        Log a dataset of input samples for NER

        :param dataset: The dataset to log. This can be an python iterable or
            Pandas/Vaex dataframe. If an iterable, it can be a list of elements that can
            be indexed into either via int index (tuple/list) or string/key index (dict)
        :param batch_size: Number of samples to log in a batch. Default 100,000
        :param text: The key/index of the text fields
        :param id: The key/index of the id fields
        :param text_token_indices: The key/index of the sample text_token_indices
        :param gold_spans: The key/index of the sample gold_spans
        :param split: train/test/validation/inference. Can be set here or via
            dq.set_split
        :param meta: List[str, int]: The keys/indexes of each metadata field.
            Consider a pandas dataframe, this would be the list of columns corresponding
            to each metadata field to log
        """
        self.validate_kwargs(kwargs)
        self.split = split
        meta = meta or []
        column_map = {
            text: "text",
            id: "id",
            gold_spans: "gold_spans",
            text_token_indices: "text_token_indices",
        }
        if isinstance(dataset, pd.DataFrame):
            dataset = dataset.rename(columns=column_map)
            self._log_df(dataset, meta)
        elif isinstance(dataset, DataFrame):
            for chunk in range(0, len(dataset), batch_size):
                chunk_df = dataset[chunk : chunk + batch_size]
                chunk_df = rename_df(chunk_df, column_map)
                self._log_df(chunk_df, meta)
        elif self.is_hf_dataset(dataset):
            self._log_hf_dataset(
                dataset,
                batch_size,
                text,
                id,
                text_token_indices,
                gold_spans,
                meta,
                split,
            )
        elif isinstance(dataset, Iterable):
            self._log_iterator(
                dataset,
                batch_size,
                text,
                id,
                text_token_indices,
                gold_spans,
                meta,
                split,
            )
        else:
            raise GalileoException(
                f"Dataset must be one of pandas, vaex, or Iterable, "
                f"but got {type(dataset)}"
            )

    def _log_hf_dataset(
        self,
        dataset: Any,
        batch_size: int,
        text: Union[str, int],
        id: Union[str, int],
        text_token_indices: Union[str, int],
        gold_spans: Union[str, int],
        meta: List[Union[str, int]],
        split: Optional[Split] = None,
    ) -> None:
        """Helper function to log a huggingface dataset

        HuggingFace datasets can be sliced, returning a dict that is in the correct
        format to log directly.
        """
        for i in range(0, len(dataset), batch_size):
            chunk = dataset[i : i + batch_size]
            chunk_meta: Dict = {col: chunk[col] for col in meta}
            self.log_data_samples(
                texts=chunk[text],
                ids=chunk[id],
                text_token_indices=chunk[text_token_indices],
                gold_spans=chunk[gold_spans],
                split=split,
                meta=chunk_meta,
            )

    def _log_iterator(
        self,
        dataset: Iterable,
        batch_size: int,
        text: Union[str, int],
        id: Union[str, int],
        text_token_indices: Union[str, int],
        gold_spans: Union[str, int],
        meta: List[Union[str, int]],
        split: Optional[Split] = None,
    ) -> None:
        batches = defaultdict(list)
        metas = defaultdict(list)
        for chunk in dataset:
            batches["text"].append(chunk[text])
            batches["gold_spans"].append(chunk[gold_spans])
            batches["text_token_indices"].append(chunk[text_token_indices])
            batches["id"].append(chunk[id])

            for meta_col in meta:
                metas[meta_col].append(self._convert_tensor_to_py(chunk[meta_col]))

            if len(batches["text"]) >= batch_size:
                self._log_dict(batches, metas, split)
                batches.clear()
                metas.clear()

        # in case there are any left
        if batches:
            self._log_dict(batches, metas, split)

    def _log_dict(
        self,
        d: Dict,
        meta: Dict,
        split: Optional[Split] = None,
    ) -> None:
        self.log_data_samples(
            texts=d["text"],
            ids=d["id"],
            text_token_indices=d["text_token_indices"],
            gold_spans=d["gold_spans"],
            split=split,
            meta=meta,
        )

    def _log_df(
        self, df: Union[pd.DataFrame, DataFrame], meta: List[Union[str, int]]
    ) -> None:
        """Helper to log a pandas or vaex df"""
        self.texts = df["text"].tolist()
        self.ids = df["id"].tolist()
        self.text_token_indices = df["text_token_indices"].tolist()
        self.gold_spans = df["gold_spans"].tolist()
        for meta_col in meta:
            self.meta[str(meta_col)] = df[meta_col].tolist()
        self.log()

    def validate(self) -> None:
        """
        Validates that the current config is correct.
        * Text and Labels must both exist (unless split is 'inference' in which case
        labels must be None)
        * Text and Labels must be the same length
        * If ids exist, it must be the same length as text/labels
        :return: None
        """
        super().validate()
        assert self.logger_config.labels, (
            "You must set your labels before logging input data. "
            "See dataquality.set_labels_for_run"
        )

        assert self.logger_config.tagging_schema, (
            "You must set your tagging schema before logging input data. "
            "See dataquality.set_tagging_schema"
        )

        text_tokenized_len = len(self.text_token_indices)
        text_len = len(self.texts)
        gold_span_len = len(self.gold_spans)
        id_len = len(self.ids)

        assert id_len == text_len, (
            f"Ids exists but are not the same length as text and labels. "
            f"(ids, text) ({id_len}, {text_len})"
        )

        if self.split == Split.inference.value:
            assert not gold_span_len, "You cannot have labels in your inference split!"
        else:
            assert gold_span_len and text_len, (
                f"You must log both text and gold_spans for split {self.split}."
                f" Text samples logged:{text_len}, gold spans logged:{gold_span_len}"
            )

            assert text_len == text_tokenized_len == gold_span_len, (
                f"gold spans, text, and tokenized text must be the same length for "
                f"split {self.split}, but got (gold spans, text, text_token) "
                f"({gold_span_len},{text_len},{text_tokenized_len})"
            )

        for sample_id, sample_spans, sample_indices, sample_text in zip(
            self.ids, self.gold_spans, self.text_token_indices, self.texts
        ):
            self._validate_sample_spans(sample_spans, sample_indices, sample_text)

            updated_spans = self._extract_gold_spans(sample_spans, sample_indices)

            sample_key = self.logger_config.get_sample_key(Split(self.split), sample_id)
            self.logger_config.gold_spans[sample_key] = [
                (span["start"], span["end"], span["label"]) for span in updated_spans
            ]
            # Unpadded length of the sample. Used to extract true predicted spans
            # which are padded by the model
            self.logger_config.sample_length[sample_key] = len(sample_indices)
            # Flatten the List[Tuple[int,int]] to List[int]
            flattened_indices = list(itertools.chain(*sample_indices))
            self.text_token_indices_flat.append(flattened_indices)

        # Free up the memory, we don't need it anymore
        del self.text_token_indices

        self.validate_metadata(batch_size=text_len)

    def _validate_sample_spans(
        self,
        sample_spans: List[Dict],
        sample_indices: List[Tuple[int, int]],
        sample_text: str,
    ) -> None:
        """Validates spans of a sample"""
        clean_labels = self._clean_labels()
        max_end_idx, max_start_idx = 0, 0
        for span in sample_spans:
            assert span["label"] in clean_labels, (
                f"'{span['label']}' is absent in the provided labels {clean_labels}. "
                f"Set all labels for run with `dataquality.set_labels_for_run`"
            )

            assert isinstance(span, dict), "individual spans must be dictionaries"
            assert "start" in span and "end" in span and "label" in span, (
                "gold_spans must have a 'start', 'end', and 'label', but got "
                f"{span.keys()}"
            )
            assert (
                span["start"] < span["end"]
            ), f"end index must be >= start index, but got {span}"
            max_end_idx = max(span["end"], max_end_idx)
            max_start_idx = max(span["start"], max_start_idx)

        assert max_start_idx <= sample_indices[-1][0], (
            f"span start idx: {max_start_idx}, does not align with provided token "
            f"boundaries {sample_indices}"
        )
        assert max_end_idx <= sample_indices[-1][1], (
            f"span end idx: {max_end_idx}, does not align with provided token "
            f"boundaries {sample_indices}"
        )
        assert max_end_idx <= len(sample_text), (
            f"span end idx: {max_end_idx} "
            f"overshoots text length of {len(sample_text)} {sample_text}"
        )

    def _extract_gold_spans(
        self, gold_spans: List[Dict], token_indices: List[Tuple[int, int]]
    ) -> List[Dict]:
        """
        This function converts gold spans that were character indexed into gold spans
            that are token indexed.
        This is done to align with the predicted spans of the model, and compute DEP
        gold_spans = [{'start': 17, 'end': 29, 'label': 'ACTOR'}]
        token_indices = [
            (0, 4), (5, 11), (12, 16), (17, 22), (17, 22), (23, 29), (23, 29)
            ]
        new_gold_spans =  [{'start': 3, 'end': 7, 'label': 'ACTOR'}]
        """
        new_gold_spans: List[Dict] = []
        for span in gold_spans:
            start = span["start"]
            end = span["end"]
            new_start, new_end = None, None
            for token_idx, token in enumerate(token_indices):
                token_start, token_end = token
                if start == token_start and new_start is None:
                    new_start = token_idx
                if end == token_end:
                    new_end = token_idx + 1
            if (
                new_start is not None and new_end is not None
            ):  # Handle edge case of where sentence > allowed_max_length
                new_gold_spans.append(
                    {
                        "start": new_start,
                        "end": new_end,
                        "label": span["label"],
                    }
                )
        assert len(new_gold_spans) == len(gold_spans), (
            f"error in span alignment, "
            f"cannot find all gold spans: "
            f"{gold_spans} in token boundaries: {token_indices}"
        )
        return new_gold_spans

    def _get_input_df(self) -> DataFrame:
        """NER is a special case where we need to log 2 different input data files

        The first is at the sentence level (id, split, text, **meta)
        The second is at the span level. This is because each sentence will have
        an arbitrary number of spans so we wont be able to create a structured
        dataframe (column numbers wont align).

        So the span-level dataframe is a row per span, with a 'sentence_id' linking
        back to the sentence.

        This function will be used for the sentence level, as that enables the parent's
        `log()` function to behave exactly as expected.
        """
        df_len = len(self.texts)
        inp = dict(
            id=self.ids,
            split=[Split(self.split).value] * df_len,
            text=self.texts,
            text_token_indices=pa.array(self.text_token_indices_flat),
            data_schema_version=[__data_schema_version__] * df_len,
            **self.meta,
        )
        df = vaex.from_dict(inp)
        return df

    @classmethod
    def process_in_out_frames(
        cls,
        in_frame: DataFrame,
        out_frame: DataFrame,
        prob_only: bool,
        epoch_or_inf_name: str,
    ) -> BaseLoggerInOutFrames:
        """Processes input and output dataframes from logging

        NER is a different case where the input data is logged at the sample level,
        but output data is logged at the span level, so we need to process it
        differently

        We don't have span IDs so we don't need to validate uniqueness
        We don't join the input and output frames
        We do need to split take only the rows from in_frame from this split
        Splits the dataframes into prob, emb, and input data for uploading to minio
        """
        cls._validate_duplicate_spans(out_frame, epoch_or_inf_name)
        prob, emb, _ = cls.split_dataframe(out_frame, prob_only)
        return BaseLoggerInOutFrames(prob=prob, emb=emb, data=in_frame)

    @classmethod
    def _validate_duplicate_spans(cls, df: DataFrame, epoch_or_inf_name: str) -> None:
        """Validates that duplicate spans aren't logged for an input sample

        Duplicate spans would be spans in a sample with identical start and end spans
        """
        dup_counts = df.groupby(["sample_id", "span_start", "span_end"], agg="count")
        dup_counts = dup_counts[dup_counts["count"] > 1]
        if len(dup_counts):
            epoch_or_inf_value, split = df[[epoch_or_inf_name, "split"]][0]
            raise GalileoException(
                "It seems as though you have duplicate spans for samples in this "
                f"split. (split:{split}, {epoch_or_inf_name}:{epoch_or_inf_value}),"
                f"dups:\n {dup_counts[['sample_id', 'count']].to_records()}"
            )

    @classmethod
    def split_dataframe(
        cls, df: DataFrame, prob_only: bool
    ) -> Tuple[DataFrame, DataFrame, DataFrame]:
        """Splits the dataframe into logical grouping for minio storage

        NER is a different case, where we store the text samples as "data" and
        all of the span level data is split into only "emb" and "prob". This function
        will only return 2 modified dataframes, where the third is expected to be the
        input data logged by the user
        """
        df_copy = df.copy()
        df_copy["id"] = vaex.vrange(0, len(df_copy), dtype="int64")
        # Separate out embeddings and probabilities into their own files
        prob_cols = [
            NERCols.id.value,
            NERCols.sample_id.value,
            NERCols.split.value,
            NERCols.epoch.value,
            NERCols.is_gold.value,
            NERCols.is_pred.value,
            NERCols.span_start.value,
            NERCols.span_end.value,
            NERCols.gold.value,
            NERCols.pred.value,
            NERCols.data_error_potential.value,
            NERCols.galileo_error_type.value,
        ]
        prob = df_copy[prob_cols]
        emb_cols = (
            [NERCols.id.value] if prob_only else [NERCols.id.value, NERCols.emb.value]
        )
        emb = df_copy[emb_cols]
        return prob, emb, df_copy

    @classmethod
    def validate_labels(cls) -> None:
        """Validates and cleans labels

        For NER, labels will come with their tags and internal values, such as:
        ['[PAD]', '[CLS]', '[SEP]', 'O', 'B-ACTOR', 'I-ACTOR', 'B-YEAR', 'B-TITLE',
        'B-GENRE', 'I-GENRE', 'B-DIRECTOR', 'I-DIRECTOR', 'B-SONG', 'I-SONG', 'B-PLOT',
        'I-PLOT', 'B-REVIEW', 'B-CHARACTER', 'I-CHARACTER', 'B-RATING',
        'B-RATINGS_AVERAGE', 'I-RATINGS_AVERAGE', 'I-TITLE', 'I-RATING', 'B-TRAILER',
        'I-TRAILER', 'I-REVIEW', 'I-YEAR']

        But we want only the true tag values (preserving order):
        ['ACTOR', 'YEAR', 'TITLE', 'GENRE', 'DIRECTOR', 'SONG', 'PLOT', 'REVIEW',
        'CHARACTER', 'RATING', 'RATINGS_AVERAGE', 'TRAILER']
        """
        cls.logger_config.labels = cls._clean_labels()

    @classmethod
    def _clean_labels(cls) -> List[str]:
        assert cls.logger_config.labels, (
            "You must set your config labels before calling finish. "
            "See `dataquality.set_labels_for_run`"
        )

        clean_labels = []
        for i in cls.logger_config.labels:
            if i and cls.is_valid_span_label(i):
                label = i.split("-", maxsplit=1)[1]
                assert len(label) != 0, (
                    "The class names following the tag cannot be empty. For example "
                    "'B-' is not allowed, but 'B-some_class' is."
                )
                clean_labels.append(label)

        assert len(clean_labels) != 0, (
            f"No valid labels found among {cls.logger_config.labels}. A valid label "
            f"is one that starts with either B-, I-, L-, E-, S-, or U- according to "
            f"your particular tagging scheme"
        )

        clean_labels = list(dict.fromkeys(clean_labels))  # Remove dups, keep order
        return clean_labels

    @classmethod
    def is_valid_span_label(cls, label: str) -> bool:
        """Denotes if a span label is valid based on our allowed tagging schemas

        B = Before the sequence
        I = In the sequence
        L/E = Last/Ending character of the sequence
        S/U = Single/Unit element of a sequence

        A valid span label would then start with 'B-' for example.
        """
        return (
            label.startswith("B-")
            or label.startswith("I-")
            or label.startswith("L-")
            or label.startswith("E-")
            or label.startswith("S-")
            or label.startswith("U-")
        )

    @classmethod
    def set_tagging_schema(cls, tagging_schema: TaggingSchema) -> None:
        cls.logger_config.tagging_schema = tagging_schema



from typing import Dict, List, Optional

from pydantic import BaseModel, StrictStr, root_validator


class MetaFilter(BaseModel):
    """A class for filtering arbitrary metadata dataframe columns

    For example, to filter on a logged metadata column, "is_happy" for values [True],
    you can create a MetaFilter(name="is_happy", isin=[True])

    You can use this filter for any columns, not just metadata columns.
    For example, you can use this to filter for DEP scores above 0.5:
    MetaFilter(name="data_error_potential", greater_than=0.5)
    """

    name: StrictStr
    greater_than: Optional[float] = None
    less_than: Optional[float] = None
    isin: Optional[List[str]] = None


class InferenceFilter(BaseModel):
    """A class for filtering an inference split

    - `is_otb`: Filters samples that are / are not On-The-Boundary
    - `is_drifted`: Filters samples that are / are not Drifted
    """

    is_otb: Optional[bool] = None
    is_drifted: Optional[bool] = None


class LassoSelection(BaseModel):
    """Representation of a lasso selection (used during an embeddings selection)

    x and y correspond to the cursor movement while tracing the lasso. This is natively
    provided by plotly when creating a lasso selection
    """

    x: List[float]
    y: List[float]

    @root_validator()
    def validate_xy(cls: BaseModel, values: Dict[str, List]) -> Dict[str, List]:
        if len(values.get("x", [])) != len(values.get("y", [])):
            raise ValueError("x and y must have the same number of points")
        if len(values.get("x", [])) < 1:
            raise ValueError("x and y must have at least 1 value")
        return values


class FilterParams(BaseModel):
    """A class for sending filters to the API alongside most any request.

    Each field represents things you can filter the dataframe on.

    Args:
        ids: List[int] = []  filter for specific IDs in the dataframe (span IDs for NER)
        similar_to: Optional[int] = None  provide an ID to run similarity search on
        num_similar_to: Optional[int] = None  if running similarity search, how many
        text_pat: Optional[StrictStr] = None  filter text samples by some text pattern
        regex: Optional[bool] = None  if searching with text, whether to use regex
        data_error_potential_high: Optional[float] = None  only samples with DEP <= this
        data_error_potential_low: Optional[float] = None  only samples with DEP >= this
        misclassified_only: Optional[bool] = None  Only look at missed samples
        gold_filter: Optional[List[StrictStr]] = None  filter GT classes
        pred_filter: Optional[List[StrictStr]] = None  filter prediction classes
        meta_filter: Optional[List[MetaFilter]] = None  see MetaFilter class
        inference_filter: Optional[InferenceFilter] = None  see InferenceFilter class
        span_sample_ids: Optional[List[int]] = None  (NER only) filter for full samples
        span_text: Optional[str] = None  (NER only) filter only on span text
        exclude_ids: List[int] = []  opposite of `ids`
        lasso: Optional[LassoSelection] = None  see LassoSelection class
        class_filter: Optional[List[StrictStr]] = None  filter GT OR prediction
    """

    ids: List[int] = []
    similar_to: Optional[List[int]] = None
    num_similar_to: Optional[int] = None
    text_pat: Optional[StrictStr] = None
    regex: Optional[bool] = None
    data_error_potential_high: Optional[float] = None
    data_error_potential_low: Optional[float] = None
    misclassified_only: Optional[bool] = None
    gold_filter: Optional[List[StrictStr]] = None
    pred_filter: Optional[List[StrictStr]] = None
    meta_filter: Optional[List[MetaFilter]] = None
    inference_filter: Optional[InferenceFilter] = None
    span_sample_ids: Optional[List[int]] = None
    span_text: Optional[str] = None
    span_regex: Optional[bool] = None
    exclude_ids: List[int] = []
    lasso: Optional[LassoSelection] = None
    class_filter: Optional[List[StrictStr]] = None
    likely_mislabeled: Optional[bool] = None



from typing import Any, Dict, List, Optional, Type, Union

import numpy as np

from dataquality.core._config import config
from dataquality.exceptions import GalileoException
from dataquality.loggers.data_logger import BaseGalileoDataLogger
from dataquality.loggers.data_logger.base_data_logger import ITER_CHUNK_SIZE, DataSet
from dataquality.loggers.logger_config.text_multi_label import (
    text_multi_label_logger_config,
)
from dataquality.loggers.model_logger import BaseGalileoModelLogger
from dataquality.schemas.ner import TaggingSchema
from dataquality.schemas.split import Split
from dataquality.schemas.task_type import TaskType
from dataquality.utils.helpers import check_noop

DEFAULT_RANDOM_EMB_DIM = 2


@check_noop
def log_data_samples(
    *,
    texts: List[str],
    ids: List[int],
    meta: Dict[str, List[Union[str, float, int]]] = None,
    **kwargs: Any,
) -> None:
    """Logs a batch of input samples for model training/test/validation/inference.

    Fields are expected as lists of their content. Field names are in the plural of
    `log_input_sample` (text -> texts)
    The expected arguments come from the task_type being used: See dq.docs() for details

    ex (text classification):
    .. code-block:: python

        all_labels = ["A", "B", "C"]
        dq.set_labels_for_run(labels = all_labels)

        texts: List[str] = [
            "Text sample 1",
            "Text sample 2",
            "Text sample 3",
            "Text sample 4"
        ]

        labels: List[str] = ["B", "C", "A", "A"]

        meta = {
            "sample_importance": ["high", "low", "low", "medium"]
            "quality_ranking": [9.7, 2.4, 5.5, 1.2]
        }

        ids: List[int] = [0, 1, 2, 3]
        split = "training"

        dq.log_data_samples(texts=texts, labels=labels, ids=ids, meta=meta split=split)

    :param texts: List[str] the input samples to your model
    :param ids: List[int | str] the ids per sample
    :param split: Optional[str] the split for this data. Can also be set via
        dq.set_split
    :param meta: Dict[str, List[str | int | float]]. Log additional metadata fields to
    each sample. The name of the field is the key of the dictionary, and the values are
    a list that correspond in length and order to the text samples.
    :param kwargs: See dq.docs() for details on other task specific parameters
    """
    assert all(
        [config.task_type, config.current_project_id, config.current_run_id]
    ), "You must call dataquality.init before logging data"
    data_logger = get_data_logger()
    data_logger.log_data_samples(texts=texts, ids=ids, meta=meta, **kwargs)


@check_noop
def log_data_sample(*, text: str, id: int, **kwargs: Any) -> None:
    """Log a single input example to disk

    Fields are expected singular elements. Field names are in the singular of
    `log_input_samples` (texts -> text)
    The expected arguments come from the task_type being used: See dq.docs() for details

    :param text: List[str] the input samples to your model
    :param id: List[int | str] the ids per sample
    :param split: Optional[str] the split for this data. Can also be set via
        dq.set_split
    :param kwargs: See dq.docs() for details on other task specific parameters
    """
    assert all(
        [config.task_type, config.current_project_id, config.current_run_id]
    ), "You must call dataquality.init before logging data"
    data_logger = get_data_logger()
    # For logging a single sample, we don't want a progress bar because it will be
    # nearly instant, and it's likely that the user will call this many times which
    # would flood the output.
    # We don't need to reset log_export_progress because this class instance is
    # ephemeral
    data_logger.log_export_progress = False
    data_logger.log_data_sample(text=text, id=id, **kwargs)


@check_noop
def log_dataset(
    dataset: DataSet,
    *,
    batch_size: int = ITER_CHUNK_SIZE,
    text: Union[str, int] = "text",
    id: Union[str, int] = "id",
    split: Optional[Split] = None,
    meta: Optional[List[Union[str, int]]] = None,
    **kwargs: Any,
) -> None:
    """Log an iterable or other dataset to disk. Useful for logging memory mapped files

    Dataset provided must be an iterable that can be traversed row by row, and for each
    row, the fields can be indexed into either via string keys or int indexes. Pandas
    and Vaex dataframes are also allowed, as well as HuggingFace Datasets

    valid examples:
        d = [
            {"my_text": "sample1", "my_labels": "A", "my_id": 1, "sample_quality": 5.3},
            {"my_text": "sample2", "my_labels": "A", "my_id": 2, "sample_quality": 9.1},
            {"my_text": "sample3", "my_labels": "B", "my_id": 3, "sample_quality": 2.7},
        ]
        dq.log_dataset(
            d, text="my_text", id="my_id", label="my_labels", meta=["sample_quality"]
        )

        Logging a pandas dataframe, df:
              text label  id  sample_quality
        0  sample1     A   1             5.3
        1  sample2     A   2             9.1
        2  sample3     B   3             2.7
        # We don't need to set text id or label because it matches the default
        dq.log_dataset(d, meta=["sample_quality"])

        Logging and iterable of tuples:
        d = [
            ("sample1", "A", "ID1"),
            ("sample2", "A", "ID2"),
            ("sample3", "B", "ID3"),
        ]
        dq.log_dataset(d, text=0, id=2, label=1)

    Invalid example:
        d = {
            "my_text": ["sample1", "sample2", "sample3"],
            "my_labels": ["A", "A", "B"],
            "my_id": [1, 2, 3],
            "sample_quality": [5.3, 9.1, 2.7]
        }

    In the invalid case, use `dq.log_data_samples`:
        meta = {"sample_quality": d["sample_quality"]}
        dq.log_data_samples(
            texts=d["my_text"], labels=d["my_labels"], ids=d["my_ids"], meta=meta
        )

    Keyword arguments are specific to the task type. See dq.docs() for details

    :param dataset: The iterable or dataframe to log
    :batch_size: The number of data samples to log at a time. Useful when logging a
    memory mapped dataset. A larger batch_size will result in faster logging at the
    expense of more memory usage. Default 100,000
    :param text: str | int The column, key, or int index for text data. Default "text"
    :param id: str | int The column, key, or int index for id data. Default "id"
    :param split: Optional[str] the split for this data. Can also be set via
        dq.set_split
    :param meta: List[str | int] Additional keys/columns to your input data to be
        logged as metadata. Consider a pandas dataframe, this would be the list of
        columns corresponding to each metadata field to log
    :param kwargs: See help(dq.get_data_logger().log_dataset) for more details here
    or dq.docs() for more general task details
    """
    assert all(
        [config.task_type, config.current_project_id, config.current_run_id]
    ), "You must call dataquality.init before logging data"
    data_logger = get_data_logger()
    data_logger.log_dataset(
        dataset,
        batch_size=batch_size,
        text=text,
        id=id,
        split=split,
        meta=meta,
        **kwargs,
    )


@check_noop
def log_model_outputs(
    *,
    embs: Optional[Union[List, np.ndarray]],
    ids: Union[List, np.ndarray],
    split: Optional[Split] = None,
    epoch: Optional[int] = None,
    logits: Union[List, np.ndarray] = None,
    probs: Union[List, np.ndarray] = None,
    inference_name: str = None,
    exclude_embs: bool = False,
) -> None:
    """Logs model outputs for model during training/test/validation.

    :param embs: The embeddings per output sample
    :param ids: The ids for each sample. Must match input ids of logged samples
    :param split: The current split. Must be set either here or via dq.set_split
    :param epoch: The current epoch. Must be set either here or via dq.set_epoch
    :param logits: The logits for each sample
    :param probs: Deprecated, use logits. If passed in, a softmax will NOT be applied
    :param inference_name: Inference name indicator for this inference split.
        If logging for an inference split, this is required.
    :param exclude_embs: Optional flag to exclude embeddings from logging. If True and
        embs is set to None, this will generate random embs for each sample.

    The expected argument shapes come from the task_type being used
    See dq.docs() for more task specific details on parameter shape
    """
    assert all(
        [config.task_type, config.current_project_id, config.current_run_id]
    ), "You must call dataquality.init before logging data"
    assert (probs is not None) or (
        logits is not None
    ), "You must provide either logits or probs"
    assert (embs is None and exclude_embs) or (
        embs is not None and not exclude_embs
    ), "embs can be None if and only if exclude_embs is True"
    if embs is None and exclude_embs:
        embs = np.random.rand(len(ids), DEFAULT_RANDOM_EMB_DIM)
    model_logger = get_model_logger()(
        embs=embs,
        ids=ids,
        split=Split[split].value if split else "",
        epoch=epoch,
        logits=logits,
        probs=probs,
        inference_name=inference_name,
    )
    model_logger.log()


@check_noop
def set_labels_for_run(labels: Union[List[List[str]], List[str]]) -> None:
    """
    Creates the mapping of the labels for the model to their respective indexes.

    :param labels: An ordered list of labels (ie ['dog','cat','fish']
    If this is a multi-label type, then labels are a list of lists where each inner
    list indicates the label for the given task

    This order MUST match the order of probabilities that the model outputs.

    In the multi-label case, the outer order (order of the tasks) must match the
    task-order of the task-probabilities logged as well.
    """
    get_data_logger().logger_config.labels = labels


@check_noop
def set_tasks_for_run(tasks: List[str], binary: bool = False) -> None:
    """Sets the task names for the run (multi-label case only).

    This order MUST match the order of the labels list provided in log_input_data
    and the order of the probability vectors provided in log_model_outputs.

    This also must match the order of the labels logged in set_labels_for_run (meaning
    that the first list of labels must be the labels of the first task passed in here)

    :param tasks: The list of tasks for your run
    :param binary: Whether this is a binary multi label run. If true, tasks will also
    be set as your labels, and you should NOT call `dq.set_labels_for_run` it will be
    handled for you. Default False
    """
    if config.task_type != TaskType.text_multi_label:
        raise GalileoException("You can only set task names for multi-label use cases.")
    get_data_logger().logger_config.tasks = tasks
    text_multi_label_logger_config.binary = binary
    if binary:
        # The labels validator will handle adding the "NOT_" to each label
        text_multi_label_logger_config.labels = [[task] for task in tasks]


@check_noop
def set_tagging_schema(tagging_schema: TaggingSchema) -> None:
    """Sets the tagging schema for NER models

    Only valid for text_ner task_types. Others will throw an exception
    """
    get_data_logger().set_tagging_schema(tagging_schema)


def get_model_logger(task_type: TaskType = None) -> Type[BaseGalileoModelLogger]:
    task_type = _get_task_type(task_type)
    return BaseGalileoModelLogger.get_logger(task_type)


def get_data_logger(task_type: TaskType = None) -> BaseGalileoDataLogger:
    task_type = _get_task_type(task_type)
    return BaseGalileoDataLogger.get_logger(task_type)()


def _get_task_type(task_type: TaskType = None) -> TaskType:
    task = task_type or config.task_type
    if not task:
        raise GalileoException(
            "You must provide either a task_type or first call"
            "dataqualtiy.init and provide one"
        )
    return task


def docs() -> None:
    """Print the documentation for your specific input and output logging format

    Based on your task_type, this will print the appropriate documentation
    """
    get_data_logger().doc()
    get_model_logger().doc()


@check_noop
def set_epoch(epoch: int) -> None:
    """Set the current epoch.

    When set, logging model outputs will use this if not logged explicitly
    """
    get_data_logger().logger_config.cur_epoch = epoch


@check_noop
def set_split(split: Split, inference_name: Optional[str] = None) -> None:
    """Set the current split.

    When set, logging data inputs/model outputs will use this if not logged explicitly
    When setting split to inference, inference_name must be included
    """
    get_data_logger().logger_config.cur_inference_name = inference_name
    split = Split[split]
    setattr(get_data_logger().logger_config, f"{split}_logged", True)
    # Set cur_inference_name before split for pydantic validation
    get_data_logger().logger_config.cur_split = split



from collections import defaultdict
from enum import Enum, unique
from typing import Any, DefaultDict, Dict, Iterable, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import vaex
from vaex.dataframe import DataFrame

from dataquality.exceptions import GalileoException
from dataquality.loggers.data_logger.base_data_logger import (
    ITER_CHUNK_SIZE,
    BaseGalileoDataLogger,
    DataSet,
    MetasType,
    MetaType,
)
from dataquality.loggers.logger_config.text_classification import (
    text_classification_logger_config,
)
from dataquality.schemas import __data_schema_version__
from dataquality.schemas.split import Split
from dataquality.utils.vaex import rename_df


@unique
class GalileoDataLoggerAttributes(str, Enum):
    texts = "texts"
    labels = "labels"
    ids = "ids"
    # mixin restriction on str (due to "str".split(...))
    split = "split"  # type: ignore
    meta = "meta"  # Metadata columns for logging
    inference_name = "inference_name"

    @staticmethod
    def get_valid() -> List[str]:
        return list(map(lambda x: x.value, GalileoDataLoggerAttributes))


class TextClassificationDataLogger(BaseGalileoDataLogger):
    """
    Class for logging input data/metadata of Text Classification models to Galileo.

    * texts: The raw text inputs for model training. List[str]
    * labels: the ground truth labels aligned to each text field. List[str]
    * ids: Optional unique indexes for each record. If not provided, will default to
    the index of the record. Optional[List[int]]
    * split: The split for training/test/validation

    ex:
    .. code-block:: python

        all_labels = ["A", "B", "C"]
        dq.set_labels_for_run(labels = all_labels)

        texts: List[str] = [
            "Text sample 1",
            "Text sample 2",
            "Text sample 3",
            "Text sample 4"
        ]

        labels: List[str] = ["B", "C", "A", "A"]

        ids: List[int] = [0, 1, 2, 3]
        meta = {"sample_quality": [5.3, 9.1, 2.7, 5.8]}
        split = "training"

        dq.log_data_samples(texts=texts, labels=labels, ids=ids, meta=meta, split=split)
    """

    __logger_name__ = "text_classification"
    logger_config = text_classification_logger_config

    def __init__(
        self,
        texts: List[str] = None,
        labels: List[str] = None,
        ids: List[int] = None,
        split: str = None,
        meta: MetasType = None,
        inference_name: str = None,
    ) -> None:
        """Create data logger.

        :param texts: The raw text inputs for model training. List[str]
        :param labels: the ground truth labels aligned to each text field.
        List[str]
        :param ids: Optional unique indexes for each record. If not provided, will
        default to the index of the record. Optional[List[Union[int,str]]]
        :param split: The split for training/test/validation
        """
        super().__init__(meta)
        # Need to compare to None because they may be np arrays which cannot be
        # evaluated with bool directly
        self.texts = texts if texts is not None else []
        self.labels = [str(i) for i in labels] if labels is not None else []
        self.ids = ids if ids is not None else []
        self.split = split
        self.inference_name = inference_name

    def log_data_samples(
        self,
        *,
        texts: List[str],
        ids: List[int],
        labels: Optional[List[str]] = None,
        split: Optional[Split] = None,
        inference_name: Optional[str] = None,
        meta: Optional[MetasType] = None,
        **kwargs: Any,  # For typing
    ) -> None:
        """Log input samples for text classification

        ex:
        .. code-block:: python

            dq.init("text_classification")
            all_labels = ["A", "B", "C"]
            dq.set_labels_for_run(labels = all_labels)

            texts: List[str] = [
                "Text sample 1",
                "Text sample 2",
                "Text sample 3",
                "Text sample 4"
            ]

            labels: List[str] = ["B", "C", "A", "A"]

            ids: List[int] = [0, 1, 2, 3]
            split = "training"

            dq.log_data_samples(texts=texts, labels=labels, ids=ids, split=split)

        :param texts: List[str] text samples
        :param ids: List[int | str] IDs for each text sample
        :param labels: List[str] labels for each text sample.
            Required if not in inference
        :param split: train/test/validation/inference. Can be set here or via
            dq.set_split
        :param inference_name: If logging inference data, a name for this inference
            data is required. Can be set here or via dq.set_split
        :param meta: Dict[str, List[str, int, float]]. Metadata for each text sample
            Format is the {"metadata_field_name": [metdata value per sample]}
        """
        self.validate_kwargs(kwargs)
        self.texts = texts
        self.ids = ids
        self.split = split
        self.labels = [str(i) for i in labels] if labels is not None else []
        self.inference_name = inference_name
        self.meta = meta or {}
        self.log()

    def log_data_sample(
        self,
        *,
        text: str,
        id: int,
        label: Optional[str] = None,
        split: Optional[Split] = None,
        inference_name: Optional[str] = None,
        meta: Optional[MetaType] = None,
        **kwargs: Any,
    ) -> None:
        """
        Log a single input sample for text classification

        :param text: str the text sample
        :param id: The sample ID
        :param label: str label for the sample. Required if not in inference
        :param split: train/test/validation/inference. Can be set here or via
            dq.set_split
        :param inference_name: If logging inference data, a name for this inference
            data is required. Can be set here or via dq.set_split
        :param meta: Dict[str, Union[str, int, float]]. Metadata for the text sample
            Format is the {"metadata_field_name": metadata_field_value}
        """
        self.validate_kwargs(kwargs)
        self.texts = [text]
        self.ids = [id]
        self.split = split
        self.labels = [label] if label else []
        self.inference_name = inference_name
        self.meta = {i: [meta[i]] for i in meta} if meta else {}
        self.log()

    def log_dataset(
        self,
        dataset: DataSet,
        *,
        batch_size: int = ITER_CHUNK_SIZE,
        text: Union[str, int] = "text",
        id: Union[str, int] = "id",
        label: Optional[Union[str, int]] = "label",
        split: Optional[Split] = None,
        inference_name: Optional[str] = None,
        meta: Optional[List[Union[str, int]]] = None,
        **kwargs: Any,
    ) -> None:
        """
        Log a dataset of input samples for text classification

        :param dataset: The dataset to log. This can be an python iterable or
            Pandas/Vaex dataframe. If an iterable, it can be a list of elements that can
            be indexed into either via int index (tuple/list) or string/key index (dict)
        :param batch_size: Number of samples to log in a batch. Default 100,000
        :param text: The key/index of the text fields
        :param id: The key/index of the id fields
        :param label: The key/index of the label fields
        :param split: train/test/validation/inference. Can be set here or via
            dq.set_split
        :param inference_name: If logging inference data, a name for this inference
            data is required. Can be set here or via dq.set_split
        :param meta: List[str, int]: The keys/indexes of each metadata field.
            Consider a pandas dataframe, this would be the list of columns corresponding
            to each metadata field to log
        """
        self.validate_kwargs(kwargs)
        self.split = split
        self.inference_name = inference_name
        meta = meta or []
        column_map = {text: "text", id: "id"}
        label = None if split == Split.inference else label
        if label:
            column_map[label] = "label"
        if isinstance(dataset, pd.DataFrame):
            dataset = dataset.rename(columns=column_map)
            self._log_df(dataset, meta)
        elif isinstance(dataset, DataFrame):
            for chunk in range(0, len(dataset), batch_size):
                chunk_df = dataset[chunk : chunk + batch_size]
                chunk_df = rename_df(chunk_df, column_map)
                self._log_df(chunk_df, meta)
        elif self.is_hf_dataset(dataset):
            self._log_hf_dataset(
                dataset, batch_size, text, id, meta, label, split, inference_name
            )
        elif isinstance(dataset, Iterable):
            self._log_iterator(
                dataset, batch_size, text, id, meta, label, split, inference_name
            )
        else:
            raise GalileoException(
                f"Dataset must be one of pandas, vaex, or Iterable, "
                f"but got {type(dataset)}"
            )

    def _log_hf_dataset(
        self,
        dataset: Any,
        batch_size: int,
        text: Union[str, int],
        id: Union[str, int],
        meta: List[Union[str, int]],
        label: Union[str, int] = None,
        split: Split = None,
        inference_name: str = None,
    ) -> None:
        """Helper function to log a huggingface dataset

        HuggingFace datasets can be sliced, returning a dict that is in the correct
        format to log directly.
        """
        for i in range(0, len(dataset), batch_size):
            chunk = dataset[i : i + batch_size]
            data = dict(
                text=chunk[text], id=chunk[id], label=chunk[label] if label else None
            )
            chunk_meta = {col: chunk[col] for col in meta}
            self._log_dict(data, chunk_meta, split, inference_name)

    def _log_iterator(
        self,
        dataset: Iterable,
        batch_size: int,
        text: Union[str, int],
        id: Union[str, int],
        meta: List[Union[str, int]],
        label: Union[str, int] = None,
        split: Split = None,
        inference_name: str = None,
    ) -> None:
        batches = defaultdict(list)
        metas = defaultdict(list)
        for chunk in dataset:
            batches["text"].append(self._convert_tensor_to_py(chunk[text]))
            batches["id"].append(self._convert_tensor_to_py(chunk[id]))
            if label:
                # Process separately because multi-label needs to override this
                # to handle labels as a list of lists
                batches = self._process_label(batches, chunk[label])
            for meta_col in meta:
                metas[meta_col].append(self._convert_tensor_to_py(chunk[meta_col]))

            if len(batches["text"]) >= batch_size:
                self._log_dict(batches, metas, split, inference_name)
                batches.clear()
                metas.clear()
        # in case there are any left
        if batches:
            self._log_dict(batches, metas, split, inference_name)

    def _process_label(self, batches: DefaultDict, label: Any) -> DefaultDict:
        """Process label for text-classification and multi-label accordingly"""
        batches["label"].append(self._convert_tensor_to_py(label))
        return batches

    def _log_dict(
        self, d: Dict, meta: Dict, split: Split = None, inference_name: str = None
    ) -> None:
        self.log_data_samples(
            texts=d["text"],
            labels=d["label"],
            ids=d["id"],
            split=split,
            inference_name=inference_name,
            meta=meta,
        )

    def _log_df(
        self, df: Union[pd.DataFrame, DataFrame], meta: List[Union[str, int]]
    ) -> None:
        """Helper to log a pandas or vex df"""
        self.texts = df["text"].tolist()
        self.ids = df["id"].tolist()
        # Inference case
        if "label" in df.columns:
            self.labels = df["label"].tolist()
        for meta_col in meta:
            self.meta[str(meta_col)] = df[meta_col].tolist()
        self.log()

    @staticmethod
    def get_valid_attributes() -> List[str]:
        """
        Returns a list of valid attributes that this logger accepts
        :return: List[str]
        """
        return GalileoDataLoggerAttributes.get_valid()

    def validate(self) -> None:
        """
        Validates that the current config is correct.
        * Text and Labels must both exist (unless split is 'inference' in which case
        labels must be None)
        * Text and Labels must be the same length
        * If ids exist, it must be the same length as text/labels
        :return: None
        """
        super().validate()

        label_len = len(self.labels)
        text_len = len(self.texts)
        id_len = len(self.ids)

        self.texts = list(self._convert_tensor_ndarray(self.texts))
        clean_labels = self._convert_tensor_ndarray(self.labels, attr="Labels")
        # If the dtype if object, we have a ragged nested sequence, so we need to
        # iterate list by list converting to strings
        if clean_labels.dtype == object:
            self.labels = [np.array(i).astype("str").tolist() for i in clean_labels]
        # Normal nparray, can convert to string elements directly
        else:
            self.labels = clean_labels.astype("str").tolist()
        self.ids = list(self._convert_tensor_ndarray(self.ids))

        if self.split == Split.inference.value:
            assert not label_len, "You cannot have labels in your inference split!"
            if not self.inference_name:
                self.inference_name = self.logger_config.cur_inference_name
            assert self.inference_name, (
                "Inference name must be set when logging an inference split. Use "
                "set_split('inference', inference_name) to set inference name"
            )

        else:
            assert label_len and text_len, (
                f"You must log both text and labels for split {self.split}."
                f" Text samples logged:{text_len}, labels logged:{label_len}"
            )

            assert text_len == label_len, (
                f"labels and text must be the same length, but got"
                f"(labels, text) ({label_len}, {text_len})"
            )

        assert id_len == text_len, (
            f"Ids exists but are not the same length as text and labels. "
            f"(ids, text) ({id_len}, {text_len})"
        )

        self.validate_logged_labels()
        self.validate_metadata(batch_size=text_len)

    def validate_logged_labels(self) -> None:
        """Validates that the labels logged match the labels set"""
        self.logger_config.observed_labels.update(self.labels)
        found_labels = self.logger_config.observed_labels
        if self.logger_config.labels:
            set_labels = set(self.logger_config.labels)
            assert set_labels.issuperset(found_labels), (
                f"Labels set to {set_labels} but found logged labels {found_labels}. "
                f"Logged labels must be the same as labels set during "
                f"dq.set_labels_for_run. Fix logged data or update labels."
            )

    def _get_input_df(self) -> DataFrame:
        inp = dict(
            id=self.ids,
            text=self.texts,
            split=self.split,
            data_schema_version=__data_schema_version__,
            gold=self.labels if self.split != Split.inference.value else None,
            **self.meta,
        )
        if self.inference_name:
            inp.update(inference_name=self.inference_name)
        return vaex.from_pandas(pd.DataFrame(inp))

    @classmethod
    def split_dataframe(
        cls, df: DataFrame, prob_only: bool
    ) -> Tuple[DataFrame, DataFrame, DataFrame]:
        """Splits the singular dataframe into its 3 components

        Gets the probability df, the embedding df, and the "data" df containing
        all other columns
        """
        df_copy = df.copy()
        # Separate out embeddings and probabilities into their own files
        prob_cols = cls._get_prob_cols()
        prob = df_copy[prob_cols]

        if prob_only:  # In this case, we don't care about the other columns
            emb_cols = ["id"]
            other_cols = ["id"]
        else:
            emb_cols = ["id", "emb"]
            ignore_cols = ["emb", "split_id"] + prob_cols
            other_cols = [i for i in df_copy.get_column_names() if i not in ignore_cols]
            other_cols += ["id"]

        emb = df_copy[emb_cols]
        data_df = df_copy[other_cols]
        return prob, emb, data_df

    @classmethod
    def _get_prob_cols(cls) -> List[str]:
        return ["id", "prob", "gold"]

    @classmethod
    def validate_labels(cls) -> None:
        assert cls.logger_config.labels, (
            "You must set your config labels before calling finish. "
            "See `dataquality.set_labels_for_run`"
        )

        assert len(cls.logger_config.labels) == cls.logger_config.observed_num_labels, (
            f"You set your labels to be {cls.logger_config.labels} "
            f"({len(cls.logger_config.labels)} labels) but based on training, your "
            f"model is expecting {cls.logger_config.observed_num_labels} labels. "
            f"Use dataquality.set_labels_for_run to update your config labels."
        )

        assert cls.logger_config.observed_labels.issubset(cls.logger_config.labels), (
            f"Labels set to {cls.logger_config.labels} do not align with observed "
            f"logged labels of {cls.logger_config.observed_labels}. Set labels must "
            "contain all logged labels. Update your labels with "
            "`dq.set_labels_for_run` or fix input data."
        )



from pathlib import Path
from typing import Callable, List, Type
from unittest import mock

import datasets
import pandas as pd
import pytest
import tensorflow as tf
import vaex
from pydantic import ValidationError

import dataquality
from dataquality.exceptions import GalileoException
from dataquality.loggers.base_logger import BaseGalileoLogger, T
from dataquality.loggers.data_logger.base_data_logger import DataSet
from dataquality.loggers.data_logger.text_classification import (
    TextClassificationDataLogger,
)
from dataquality.loggers.model_logger.base_model_logger import BaseGalileoModelLogger
from dataquality.schemas.split import Split


class TestSetSplitInference:
    def test_set_split_inference(self) -> None:
        assert not dataquality.get_data_logger().logger_config.inference_logged
        dataquality.set_split("inference", "all-customers")
        assert dataquality.get_data_logger().logger_config.cur_split == "inference"
        assert (
            dataquality.get_data_logger().logger_config.cur_inference_name
            == "all-customers"
        )

    def test_set_split_inference_missing_inference_name(self) -> None:
        with pytest.raises(ValidationError) as e:
            dataquality.set_split("inference")

        assert (
            e.value.errors()[0]["msg"]
            == "Please specify inference_name when setting split to inference"
        )


class TestBaseLoggersInference:
    # Use _setup since setup is a protected pytest var
    def _setup(self, logger_cls: Type[T]) -> Type[T]:
        logger = logger_cls()
        logger.logger_config.reset()
        return logger

    def test_validate_inference(self) -> None:
        base_logger = self._setup(logger_cls=BaseGalileoLogger)

        assert not base_logger.split
        assert not base_logger.logger_config.inference_logged

        base_logger.logger_config.cur_inference_name = "customers"
        base_logger.logger_config.cur_split = Split.inference
        base_logger.set_split_epoch()

        assert base_logger.split
        assert base_logger.logger_config.inference_logged

    def test_base_model_logger_validate_inference(self) -> None:
        base_model_logger = self._setup(BaseGalileoModelLogger)

        assert not base_model_logger.split
        assert not base_model_logger.logger_config.inference_logged

        base_model_logger.logger_config.cur_inference_name = "customers"
        base_model_logger.logger_config.cur_split = Split.inference

        base_model_logger.set_split_epoch()
        assert base_model_logger.split
        assert base_model_logger.logger_config.inference_logged

    def test_base_model_logger_validate_inference_missing_inference_name(self) -> None:
        base_model_logger = self._setup(BaseGalileoModelLogger)

        assert not base_model_logger.split
        assert not base_model_logger.logger_config.inference_logged

        base_model_logger.split = Split.inference
        with pytest.raises(GalileoException) as e:
            base_model_logger.set_split_epoch()

        assert e.value.args[0] == (
            "For inference split you must either log an inference name or set "
            "it before logging. Use `dataquality.set_split` to setinference_name"
        )

    @mock.patch(
        "dataquality.loggers.model_logger.base_model_logger._save_hdf5_file",
        return_value="1234-abcd-5678",
    )
    @mock.patch(
        "dataquality.loggers.model_logger.base_model_logger.uuid4",
        return_value="1234-abcd-5678",
    )
    def test_write_model_output_inference(
        self, mock_uuid: mock.MagicMock, mock_save_file: mock.MagicMock
    ) -> None:
        inference_data = {
            "epoch": [None, None, None],
            "split": ["inference", "inference", "inference"],
            "inference_name": ["customers", "customers", "customers"],
        }
        logger = BaseGalileoModelLogger()
        logger.write_model_output(inference_data)

        local_file = (
            f"{Path.home()}/.galileo/logs/{dataquality.config.current_project_id}/"
            f"{dataquality.config.current_run_id}/inference/customers"
        )
        # Assert _save_hdf5_file is called with correct args
        mock_save_file.assert_called_once_with(
            local_file, "1234abcd5678.hdf5", inference_data
        )

    @pytest.mark.parametrize(
        "splits,expected",
        [
            (["training_logged"], True),
            (["test_logged"], True),
            (["validation_logged"], True),
            (["inference_logged"], False),
            (["training_logged", "inference_logged"], True),
            (["test_logged", "validation_logged"], True),
        ],
    )
    def test_non_inference_logged(self, splits: List[str], expected: bool) -> None:
        base_logger = self._setup(logger_cls=BaseGalileoLogger)
        for split in splits:
            setattr(base_logger.logger_config, split, True)

        assert base_logger.non_inference_logged() is expected


class TestTextClassificationDataLoggerInference:
    def _setup(self, **kwargs) -> TextClassificationDataLogger:
        logger = TextClassificationDataLogger(split=Split.inference, **kwargs)
        logger.logger_config.reset()
        return logger

    def test_validate_inference(self):
        logger = self._setup(**{"inference_name": "animals"})
        logger.validate()

    def test_validate_inference_with_labels(self):
        logger = self._setup(**{"labels": ["dog", "cat"], "inference_name": "animals"})
        with pytest.raises(AssertionError) as e:
            logger.validate()

        assert e.value.args[0] == "You cannot have labels in your inference split!"

    def test_validate_inference_missing_inference_name(self):
        logger = self._setup()
        with pytest.raises(AssertionError) as e:
            logger.validate()

        assert e.value.args[0] == (
            "Inference name must be set when logging an inference split. Use "
            "set_split('inference', inference_name) to set inference name"
        )

    @pytest.mark.parametrize(
        "dataset",
        [
            pd.DataFrame(
                {
                    "my_text": ["sample1", "sample2", "sample3"],
                    "my_id": [1, 2, 3],
                }
            ),
            vaex.from_dict(
                {
                    "my_text": ["sample1", "sample2", "sample3"],
                    "my_id": [1, 2, 3],
                }
            ),
            [
                {"my_text": "sample1", "my_id": 1},
                {"my_text": "sample2", "my_id": 2},
                {"my_text": "sample3", "my_id": 3},
            ],
            tf.data.Dataset.from_tensor_slices(
                {
                    "my_text": ["sample1", "sample2", "sample3"],
                    "my_id": [1, 2, 3],
                }
            ),
            datasets.Dataset.from_dict(
                dict(
                    my_text=["sample1", "sample2", "sample3"],
                    my_id=[1, 2, 3],
                )
            ),
        ],
    )
    def test_log_dataset(
        self, dataset: DataSet, set_test_config: Callable, cleanup_after_use: Callable
    ) -> None:
        set_test_config(split="inference")
        logger = TextClassificationDataLogger()

        with mock.patch("dataquality.core.log.get_data_logger") as mock_method:
            mock_method.return_value = logger
            dataquality.log_dataset(
                dataset,
                text="my_text",
                id="my_id",
                split="inference",
                inference_name="foo",
            )

            assert logger.texts == ["sample1", "sample2", "sample3"]
            assert logger.ids == [1, 2, 3]
            assert logger.split == Split.inference
            assert logger.inference_name == "foo"



import os
import time
from random import random
from typing import Callable
from unittest import mock
from unittest.mock import MagicMock
from uuid import uuid4

import numpy as np
import pytest
import vaex

import dataquality
import dataquality as dq
import dataquality.clients.api
import dataquality.core._config
import dataquality.core.finish
from dataquality.exceptions import GalileoException
from dataquality.loggers.data_logger import BaseGalileoDataLogger
from dataquality.loggers.model_logger.text_classification import (
    TextClassificationModelLogger,
)
from dataquality.schemas.task_type import TaskType
from dataquality.utils.thread_pool import ThreadPoolManager
from tests.conftest import TEST_PATH
from tests.utils.data_utils import (
    NUM_LOGS,
    NUM_RECORDS,
    _log_text_classification_data,
    validate_cleanup_data,
    validate_uploaded_data,
)

MAX_META_COLS = BaseGalileoDataLogger.MAX_META_COLS
MAX_STR_LEN = BaseGalileoDataLogger.MAX_STR_LEN


def test_threaded_logging_and_upload(
    cleanup_after_use: Callable, set_test_config: Callable
) -> None:
    """
    Tests that threaded calls to upload still yield non-missing datasets
    """
    num_records = 32
    num_logs = 20
    num_embs = 50
    _log_text_classification_data(
        num_records=num_records, num_logs=num_logs, num_embs=num_embs
    )
    try:
        # Equivalent to the users `finish` call, but we don't want to clean up files yet
        ThreadPoolManager.wait_for_threads()
        c = dataquality.get_data_logger("text_classification")
        c.validate_labels()
        c.upload()
        validate_uploaded_data(num_records * num_logs)
        c._cleanup()
        validate_cleanup_data()
    finally:
        # Mock finish() call without calling the API
        ThreadPoolManager.wait_for_threads()


def test_multi_label_logging(
    cleanup_after_use: Callable, set_test_config: Callable
) -> None:
    """
    Tests that threaded calls to upload still yield non-missing datasets
    """
    set_test_config(task_type=TaskType.text_multi_label)
    num_records = 32
    num_logs = 20
    num_embs = 50
    _log_text_classification_data(
        num_records=num_records, num_logs=num_logs, num_embs=num_embs, multi_label=True
    )
    try:
        # Equivalent to the users `finish` call, but we don't want to clean up files yet
        ThreadPoolManager.wait_for_threads()
        c = dataquality.get_data_logger()
        c.validate_labels()
        c.upload()
        validate_uploaded_data(num_records * num_logs, multi_label=True)
        c._cleanup()
        validate_cleanup_data()
    finally:
        # Mock finish() call without calling the API
        ThreadPoolManager.wait_for_threads()


def test_metadata_logging(
    cleanup_after_use: Callable, set_test_config: Callable
) -> None:
    """
    Tests that logging metadata columns persist
    """
    meta_cols = ["test1", "meta2"]
    meta = {}
    for i in meta_cols:
        meta[i] = [random() for _ in range(NUM_RECORDS * NUM_LOGS)]
    _log_text_classification_data(meta=meta)
    try:
        # Equivalent to the users `finish` call, but we don't want to clean up files yet
        ThreadPoolManager.wait_for_threads()
        c = dataquality.get_data_logger()
        c.upload()
        validate_uploaded_data(meta_cols=meta_cols)
        c._cleanup()
        validate_cleanup_data()
    finally:
        # Mock finish() call without calling the API
        ThreadPoolManager.wait_for_threads()


def test_metadata_logging_different_splits(
    cleanup_after_use: Callable, set_test_config: Callable, input_data: Callable
) -> None:
    """
    Tests that logging metadata columns only attach to the splits we log them for
    """
    training_data = input_data(meta={"training_meta": [1, 2]})
    dataquality.set_labels_for_run(training_data["labels"])
    dataquality.log_data_samples(**training_data)
    test_data = input_data(split="test", meta={"test_meta": ["foo", "bar"]})
    dataquality.log_data_samples(**test_data)

    output_data = {
        "embs": np.random.rand(2, 100),
        "logits": np.random.rand(2, 5),
        "ids": [1, 2],
        "split": "training",
        "epoch": 0,
    }
    dataquality.log_model_outputs(**output_data)
    output_data["split"] = "test"
    dataquality.log_model_outputs(**output_data)
    dataquality.get_data_logger().upload()

    train_data = vaex.open(f"{TEST_PATH}/training/0/data/data.hdf5")
    test_data = vaex.open(f"{TEST_PATH}/test/0/data/data.hdf5")

    assert "training_meta" in train_data.get_column_names()
    assert "training_meta" not in test_data.get_column_names()

    assert "test_meta" in test_data.get_column_names()
    assert "test_meta" not in train_data.get_column_names()

    assert sorted(train_data["training_meta"].tolist()) == [1, 2]
    assert sorted(test_data["test_meta"].tolist()) == ["bar", "foo"]


def test_metadata_logging_invalid(
    cleanup_after_use: Callable, set_test_config: Callable
) -> None:
    """
    Tests our metadata logging validation
    """
    meta = {
        "test1": [random() for _ in range(NUM_RECORDS * NUM_LOGS)],
        "meta2": [random() for _ in range(NUM_RECORDS * NUM_LOGS)],
        "bad_attr": [
            "te" * MAX_STR_LEN for _ in range(NUM_RECORDS * NUM_LOGS)
        ],  # String too long
        "another_bad_attr": ["test", "test", "test"],  # Wrong number of values
        # Right length, but can't contain a list
        "bad_attr_3": [[1]] + [random() for _ in range(NUM_RECORDS * NUM_LOGS - 1)],
        "gold": [random() for _ in range(NUM_RECORDS * NUM_LOGS)],  # Reserved key
    }

    # Too many metadata columns
    for i in range(MAX_META_COLS):
        meta[f"attr_{i}"] = [random() for _ in range(NUM_RECORDS * NUM_LOGS)]

    _log_text_classification_data(meta=meta)
    valid_meta_cols = ["test1", "meta2"]
    valid_meta_cols += [f"attr_{i}" for i in range(44)]
    try:
        # Equivalent to the users `finish` call, but we don't want to clean up files yet
        c = dataquality.get_data_logger("text_classification")
        c.upload()
        validate_uploaded_data(meta_cols=valid_meta_cols)
        c._cleanup()
        validate_cleanup_data()
    finally:
        # Mock finish() call without calling the API
        ThreadPoolManager.wait_for_threads()


def test_logging_duplicate_ids(
    cleanup_after_use: Callable, set_test_config: Callable
) -> None:
    """
    Tests that logging duplicate ids triggers a failure
    """
    num_records = 50
    _log_text_classification_data(num_records=num_records, unique_ids=False)
    try:
        # Equivalent to the users `finish` call, but we don't want to clean up files yet
        ThreadPoolManager.wait_for_threads()
        c = dataquality.get_data_logger("text_classification")
        with pytest.raises(GalileoException) as e:
            c.upload()

        assert str(e.value).startswith(
            "It seems as though you do not have unique ids in this split."
        )
    finally:
        # Mock finish() call without calling the API
        ThreadPoolManager.wait_for_threads()


def test_logging_inference_run(
    cleanup_after_use: Callable, set_test_config: Callable, input_data: Callable
) -> None:
    """
    Tests that logging metadata columns only attach to the splits we log them for
    """
    inference_data = input_data(
        split="inference", meta={"inference_meta_1": [3.14, 42]}
    )
    dataquality.log_data_samples(**inference_data)
    inference_data = input_data(split="inference", inference_name="last-week-customers")
    dataquality.log_data_samples(**inference_data)

    dataquality.set_split("inference", inference_name="all-customers")
    embs_1 = np.random.rand(2, 100)
    logits_1 = np.random.rand(2, 5)
    output_data = {
        "embs": embs_1,
        "logits": logits_1,
        "ids": [1, 2],
    }
    dataquality.log_model_outputs(**output_data)
    dataquality.set_split("inference", inference_name="last-week-customers")
    embs_2 = np.random.rand(2, 100)
    logits_2 = np.random.rand(2, 5)
    output_data = {
        "embs": embs_2,
        "logits": logits_2,
        "ids": [1, 2],
    }
    dataquality.log_model_outputs(**output_data)

    ThreadPoolManager.wait_for_threads()
    dataquality.get_data_logger().upload()

    inference_data_1 = vaex.open(f"{TEST_PATH}/inference/all-customers/data/data.hdf5")
    inference_data_2 = vaex.open(
        f"{TEST_PATH}/inference/last-week-customers/data/data.hdf5"
    )

    assert "inference_meta_1" in inference_data_1.get_column_names()
    assert "inference_meta_1" not in inference_data_2.get_column_names()
    assert sorted(inference_data_1["inference_meta_1"].tolist()) == [3.14, 42]

    inference_emb_1 = vaex.open(f"{TEST_PATH}/inference/all-customers/emb/emb.hdf5")
    inference_emb_2 = vaex.open(
        f"{TEST_PATH}/inference/last-week-customers/emb/emb.hdf5"
    )

    assert (inference_emb_1.emb.to_numpy() == embs_1).all()
    assert (inference_emb_2.emb.to_numpy() == embs_2).all()

    inference_prob_1 = vaex.open(f"{TEST_PATH}/inference/all-customers/prob/prob.hdf5")
    inference_prob_2 = vaex.open(
        f"{TEST_PATH}/inference/last-week-customers/prob/prob.hdf5"
    )

    assert "logits" not in inference_prob_1.get_column_names()
    assert "logits" not in inference_prob_2.get_column_names()
    assert "prob" in inference_prob_1.get_column_names()
    assert "prob" in inference_prob_2.get_column_names()

    assert (
        inference_prob_1.prob.to_numpy()
        == dataquality.get_model_logger()().convert_logits_to_probs(logits_1)
    ).all()
    assert (
        inference_prob_2.prob.to_numpy()
        == dataquality.get_model_logger()().convert_logits_to_probs(logits_2)
    ).all()


def test_logging_train_test_inference(
    cleanup_after_use: Callable, set_test_config: Callable, input_data: Callable
) -> None:
    """
    Tests that logging metadata columns only attach to the splits we log them for
    """
    dataquality.set_labels_for_run(["APPLE", "ORANGE"])
    training_data = input_data(meta={"training_meta": [1.414, 123]})
    dataquality.log_data_samples(**training_data)
    test_data = input_data(split="test", meta={"test_meta": [3.14, 42]})
    dataquality.log_data_samples(**test_data)
    inference_data = input_data(split="inference")
    dataquality.log_data_samples(**inference_data)

    dataquality.set_split("training")
    train_embs = np.random.rand(2, 100)
    train_logits = np.random.rand(2, 5)
    output_data = {
        "embs": train_embs,
        "logits": train_logits,
        "ids": [1, 2],
        "epoch": 0,
    }
    dataquality.log_model_outputs(**output_data)
    dataquality.set_split("test")
    test_embs = np.random.rand(2, 100)
    test_logits = np.random.rand(2, 5)
    output_data = {"embs": test_embs, "logits": test_logits, "ids": [1, 2], "epoch": 0}
    dataquality.log_model_outputs(**output_data)
    dataquality.set_split("inference", inference_name="all-customers")
    inf_embs = np.random.rand(2, 100)
    inf_logits = np.random.rand(2, 5)
    output_data = {
        "embs": inf_embs,
        "logits": inf_logits,
        "ids": [1, 2],
    }
    dataquality.log_model_outputs(**output_data)

    ThreadPoolManager.wait_for_threads()
    dataquality.get_data_logger().upload()

    train_data = vaex.open(f"{TEST_PATH}/training/0/data/data.hdf5")
    test_data = vaex.open(f"{TEST_PATH}/test/0/data/data.hdf5")
    inference_data = vaex.open(f"{TEST_PATH}/inference/all-customers/data/data.hdf5")

    assert "training_meta" in train_data.get_column_names()
    assert sorted(train_data["training_meta"].tolist()) == [1.414, 123]
    assert "test_meta" in test_data.get_column_names()
    assert sorted(test_data["test_meta"].tolist()) == [3.14, 42]

    train_emb_data = vaex.open(f"{TEST_PATH}/training/0/emb/emb.hdf5")
    test_emb_data = vaex.open(f"{TEST_PATH}/test/0/emb/emb.hdf5")
    inference_emb_data = vaex.open(f"{TEST_PATH}/inference/all-customers/emb/emb.hdf5")

    assert (train_emb_data.emb.to_numpy() == train_embs).all()
    assert (test_emb_data.emb.to_numpy() == test_embs).all()
    assert (inference_emb_data.emb.to_numpy() == inf_embs).all()

    train_prob_data = vaex.open(f"{TEST_PATH}/training/0/prob/prob.hdf5")
    test_prob_data = vaex.open(f"{TEST_PATH}/test/0/prob/prob.hdf5")
    inference_prob_data = vaex.open(
        f"{TEST_PATH}/inference/all-customers/prob/prob.hdf5"
    )

    assert "training_meta" in train_data.get_column_names()
    assert "training_meta" not in test_data.get_column_names()
    assert "training_meta" not in inference_data.get_column_names()

    assert "test_meta" not in train_data.get_column_names()
    assert "test_meta" in test_data.get_column_names()
    assert "test_meta" not in inference_data.get_column_names()

    assert "logits" not in train_prob_data.get_column_names()
    assert "logits" not in test_prob_data.get_column_names()
    assert "logits" not in inference_prob_data.get_column_names()

    assert (
        train_prob_data.prob.to_numpy()
        == dataquality.get_model_logger()().convert_logits_to_probs(train_logits)
    ).all()
    assert (
        test_prob_data.prob.to_numpy()
        == dataquality.get_model_logger()().convert_logits_to_probs(test_logits)
    ).all()
    assert (
        inference_prob_data.prob.to_numpy()
        == dataquality.get_model_logger()().convert_logits_to_probs(inf_logits)
    ).all()


def test_prob_only(set_test_config) -> None:
    """Tests that the last_epoch is determined at the split level"""
    logger = dataquality.get_data_logger()
    logger.logger_config.last_epoch = 5
    train_split_runs = ["0", "1", "5", "3", "4"]
    val_split_runs = ["0", "1", "3", "4"]
    test_split_runs = ["0"]

    assert logger.prob_only(train_split_runs, "training", "0")
    assert not logger.prob_only(train_split_runs, "training", "5")
    assert logger.prob_only(val_split_runs, "validation", "0")
    assert not logger.prob_only(val_split_runs, "validation", "5")
    assert not logger.prob_only(test_split_runs, "test", "0")


@mock.patch.object(dataquality.clients.api.ApiClient, "get_presigned_url")
@mock.patch.object(
    dataquality.clients.objectstore.ObjectStore, "_upload_file_from_local"
)
def test_log_invalid_model_outputs(
    mock_upload_from_local: mock.MagicMock,
    mock_presigned_url: mock.MagicMock,
    cleanup_after_use: Callable,
    set_test_config: Callable,
    input_data: Callable,
) -> None:
    """Validate that we interrupt the main process if issues occur while logging"""
    dataquality.set_labels_for_run(["APPLE", "ORANGE"])
    training_data = input_data(meta={"training_meta": [1.414, 123]})
    dataquality.log_data_samples(**training_data)

    dataquality.set_split("training")
    train_embs = np.random.rand(1, 100)  # Not enough embeddings
    train_logits = np.random.rand(2, 5)
    output_data = {
        "embs": train_embs,
        "logits": train_logits,
        "ids": [1, 2],
        "epoch": 0,
    }
    with pytest.raises(GalileoException) as e:
        mock_presigned_url.return_value = "https://google.com"
        mock_upload_from_local.return_value = None
        dataquality.log_model_outputs(**output_data)
        time.sleep(1)  # ensure the first one records a failure
        dataquality.log_model_outputs(**output_data)

    assert dataquality.get_model_logger().logger_config.exception != ""
    assert str(e.value).startswith("An issue occurred while logging model outputs.")


@mock.patch.object(dataquality.clients.api.ApiClient, "get_presigned_url")
@mock.patch.object(
    dataquality.clients.objectstore.ObjectStore, "_upload_file_from_local"
)
def test_log_invalid_model_outputs_final_thread(
    mock_upload_from_local: mock.MagicMock,
    mock_presigned_url: mock.MagicMock,
    cleanup_after_use: Callable,
    set_test_config: Callable,
    input_data: Callable,
) -> None:
    """Validate that we error on finish if issues occur while logging"""
    assert dataquality.get_model_logger().logger_config.exception == ""
    dataquality.set_labels_for_run(["APPLE", "ORANGE"])
    training_data = input_data(meta={"training_meta": [1.414, 123]})
    dataquality.log_data_samples(**training_data)

    dataquality.set_split("training")
    train_embs = np.random.rand(1, 100)  # Not enough embeddings
    train_logits = np.random.rand(2, 5)
    output_data = {
        "embs": train_embs,
        "logits": train_logits,
        "ids": [1, 2],
        "epoch": 0,
    }

    mock_presigned_url.return_value = "https://google.com"
    dataquality.log_model_outputs(**output_data)
    with pytest.raises(GalileoException) as e:
        mock_upload_from_local.return_value = None
        dataquality.get_data_logger().upload()

    assert dataquality.get_model_logger().logger_config.exception != ""
    assert str(e.value).startswith("An issue occurred while logging model outputs.")


def test_log_outputs_binary(
    set_test_config: Callable, cleanup_after_use: Callable
) -> None:
    """Tests that binary logging of logits works"""
    set_test_config(task_type="text_classification")

    logger = TextClassificationModelLogger(
        embs=np.random.rand(10, 100),  # 10 samples, 100 emb per sample
        logits=np.random.rand(10, 1),  # 10 samples
        ids=list(range(10)),
        split="training",
        epoch=0,
    )
    logger._log()

    assert not hasattr(logger, "logits")
    assert logger.probs.shape == (10, 2)  # samples X tasks X classes per task
    for sample_probs in logger.probs:
        assert np.isclose(np.sum(sample_probs), 1.0)


def test_calls_noop() -> None:
    os.environ["GALILEO_DISABLED"] = "True"
    c = dataquality.core._config.set_config()
    assert c.api_url == "http://"
    with mock.patch("dataquality.core.log.log_data_samples") as mock_log:
        dataquality.log_data_samples(texts=["test"], labels=["1"], ids=[1])
        mock_log.assert_not_called()
    del os.environ["GALILEO_DISABLED"]


@mock.patch.object(dataquality.core._config.Config, "update_file_config")
@mock.patch.object(dataquality.core.finish.os, "rename")
@mock.patch.object(dataquality.clients.api.ApiClient, "get_project_run")
@mock.patch.object(dataquality.clients.api.ApiClient, "get_project")
@mock.patch.object(dataquality.clients.api.ApiClient, "delete_run")
@mock.patch.object(dataquality.clients.api.ApiClient, "create_run")
def test_reset_run_new_task_type(
    mock_create_run: MagicMock,
    mock_delete_run: MagicMock,
    mock_get_project: MagicMock,
    mock_get_project_run: MagicMock,
    mock_rename: MagicMock,
    mock_update_file_config: MagicMock,
    set_test_config: Callable,
    cleanup_after_use: Callable,
) -> None:
    """Tests that resetting a run with a new task type changes the task type"""
    set_test_config(task_type=TaskType.text_classification)
    mock_get_project.return_value = {"name": "project_name"}
    # task_type 0 from API is text_classification
    mock_get_project_run.return_value = {"name": "run_name", "task_type": 0}
    pid = uuid4()
    rid = uuid4()
    mock_create_run.return_value = {"id": rid}
    dataquality.core.finish._reset_run(pid, rid, TaskType.text_multi_label)

    assert mock_create_run.called_once_with(pid, rid, TaskType.text_multi_label)
    assert dataquality.config.current_run_id == rid


@pytest.mark.parametrize("last_epoch", [1, 3, None, 10])
@mock.patch.object(dataquality.core.finish, "_version_check")
@mock.patch.object(dataquality.core.finish, "_reset_run")
@mock.patch.object(dataquality.core.finish, "upload_dq_log_file")
@mock.patch.object(dataquality.clients.api.ApiClient, "make_request")
@mock.patch.object(dataquality.loggers.base_logger.BaseGalileoLogger, "_cleanup")
@mock.patch.object(
    dataquality.loggers.data_logger.text_classification.TextClassificationDataLogger,
    "upload_in_out_frames",
)
def test_finish_last_epoch(
    mock_upload_frames: MagicMock,
    mock_cleanup: MagicMock,
    mock_make_request: MagicMock,
    mock_upload_log_file: MagicMock,
    mock_reset_run: MagicMock,
    mock_version_check: MagicMock,
    set_test_config: Callable,
    cleanup_after_use: Callable,
    last_epoch: int,
) -> None:
    text_inputs = ["sample1", "sample2", "sample3"] * 30
    gold = ["A", "C", "B"] * 30
    ids = list(range(90))

    dq.set_labels_for_run(["A", "B", "C"])
    dq.log_data_samples(texts=text_inputs, labels=gold, split="validation", ids=ids)
    dq.log_data_samples(texts=text_inputs, labels=gold, split="training", ids=ids)

    embs = np.random.rand(3, 100)
    logits = np.random.rand(3, 3)
    last_logged_epoch = 0
    for i in range(0, 15, 3):
        ids = list(range(i, i + 3))
        dq.log_model_outputs(
            embs=embs, logits=logits, ids=ids, split="training", epoch=i // 3
        )
        last_logged_epoch += 1
    dq.finish(last_epoch=last_epoch, wait=False)
    # if max epoch is None, all should be uploaded.
    # if max epoch is greater than the max logged, all should be uploaded
    # if max epoch is less than max logged, only log up to and including max_epoch
    max_uploaded = (
        min(last_logged_epoch, last_epoch + 1) if last_epoch else last_logged_epoch
    )
    for i, call in enumerate(mock_upload_frames.call_args_list):
        # python 3.7 vs 3.8+ compatibility
        try:  # python 3.8+
            assert int(call.args[-1]) == i
        except TypeError:  # python 3.7
            assert int(call[0][-1]) == i
    assert mock_upload_frames.call_count == max_uploaded


@mock.patch("dataquality.core.log.np.random.rand")
def test_log_model_outputs_exclude_embs(
    mock_numpy: mock.MagicMock,
    set_test_config: Callable,
    cleanup_after_use: Callable,
) -> None:
    """
    Tests that logging model outputs with missing embeddings and exclude embs
    generates random embs
    """
    # Use random.randn to avoid clash with numpy.random.rand mock
    mock_numpy.return_value = np.random.randn(10, 2)
    output_data = {
        "embs": None,
        "logits": np.random.randn(10, 5),
        "ids": list(range(1, 11)),
        "split": "training",
        "epoch": 0,
        "exclude_embs": True,
    }
    dataquality.log_model_outputs(**output_data)
    # Assert that random embeddings were created with dim 2
    mock_numpy.assert_called_once_with(10, 2)


def test_log_model_outputs_missing_embs_exclude_emb_false(
    set_test_config: Callable,
) -> None:
    """
    Tests that logging model outputs with missing embeddings raises an error
    """
    output_data = {
        "embs": None,
        "logits": np.random.rand(2, 5),
        "ids": [1, 2],
        "split": "training",
        "epoch": 0,
    }

    with pytest.raises(AssertionError) as e:
        dataquality.log_model_outputs(**output_data)
    assert str(e.value) == "embs can be None if and only if exclude_embs is True"


def test_log_model_outputs_with_embs_exclude_emb_true(
    set_test_config: Callable,
) -> None:
    """
    Tests that logging model outputs embeddings with exclude_emb raises an error
    """
    output_data = {
        "embs": np.random.rand(2, 100),
        "logits": np.random.rand(2, 5),
        "ids": [1, 2],
        "split": "training",
        "epoch": 0,
        "exclude_embs": True,
    }
    with pytest.raises(AssertionError) as e:
        dataquality.log_model_outputs(**output_data)

    assert str(e.value) == "embs can be None if and only if exclude_embs is True"



import warnings
from collections import defaultdict
from typing import Any, Callable, DefaultDict, Dict, Generator, List, Tuple, Union

import numpy as np
import thinc
from spacy.language import Language
from spacy.ml.parser_model import ParserStepModel
from spacy.ml.parser_model import precompute_hiddens as State2Vec
from spacy.pipeline._parser_internals.stateclass import StateClass
from spacy.pipeline.ner import EntityRecognizer
from spacy.tokens import Doc
from spacy.training import Example
from spacy.util import minibatch
from thinc.api import set_dropout_rate
from wrapt import CallableObjectProxy

import dataquality
from dataquality import config
from dataquality.exceptions import GalileoException
from dataquality.loggers.logger_config.text_ner import text_ner_logger_config
from dataquality.loggers.model_logger.text_ner import TextNERModelLogger
from dataquality.schemas.ner import TaggingSchema
from dataquality.schemas.split import Split, conform_split
from dataquality.utils.spacy_integration import (
    convert_spacy_ents_for_doc_to_predictions,
    convert_spacy_ner_logits_to_valid_logits,
    validate_obj,
    validate_spacy_version,
)


def log_input_examples(examples: List[Example], split: Split) -> None:
    """Logs a list of Spacy Examples using the dataquality client"""
    split = conform_split(split)
    if not dataquality.get_data_logger().logger_config.labels:
        raise GalileoException(
            "Galileo does not have any logged labels. Did you forget "
            "to call watch(nlp) before log_input_examples(...)?"
        )
    texts = []
    text_token_indices = []
    gold_spans = []
    ids = []
    for i, example in enumerate(examples):
        validate_obj(example, Example, "reference")
        # For the most part reference has all the information we want to log
        data = example.reference
        # but predicted is the Doc that will be passed along to the spacy models, and
        # crucially holds the "id" user_data we attach
        texts.append(data.text)
        text_token_indices.append(
            [(token.idx, token.idx + len(token)) for token in data]
        )
        gold_spans.append(
            [
                {"start": ent.start_char, "end": ent.end_char, "label": ent.label_}
                for ent in data.ents
            ]
        )
        # We add ids to the doc.user_data to be along for the ride through spacy
        # The predicted doc is the one that the model will see
        example.predicted.user_data["id"] = i
        ids.append(i)
    dataquality.log_data_samples(
        texts=texts,
        text_token_indices=text_token_indices,
        gold_spans=gold_spans,
        ids=ids,
        split=split,
    )


def watch(nlp: Language) -> None:
    """Stores the nlp object before calling watch on the ner component within it

    We need access to the nlp object so that during training we can capture the
    model's predictions over the raw text by running nlp("user's text") and looking
    at the results

    :param nlp: The spacy nlp Language component.
    :return: None
    """
    validate_spacy_version()
    if not (config.current_project_id and config.current_run_id):
        raise GalileoException(
            "You must initialize dataquality first! "
            "Use dataquality.init(project_name='my_cool_project', "
            "run_name='my_awesome_run', task_type='text_ner')"
        )
    # TODO: Replace with the following in the future
    # if not (config.current_project_id and config.current_run_id):
    #     dataquality.init(task_type=TaskType.text_ner)
    #     warnings.warn("No run initialized with dataquality.init(...). "
    #                   "Creating one with the project name `{}` and run_name `{}`")
    ner = nlp.get_pipe("ner")

    if "O" not in ner.move_names:  # type: ignore
        raise GalileoException(
            "Missing the 'O' tag in the model's moves, are you sure you have"
            "already called 'nlp.begin_training()' or "
            "`nlp.initialize(training_examples)`?"
        )

    if len(ner.move_names) <= 1:  # type: ignore
        raise GalileoException(
            "Your nlp seems to not have been initialized with any ground truth spans. "
            "Galileo needs all labels to have been added to the model before calling "
            "watch(nlp). Please run `nlp.initialize(lambda: your_examples)` over a "
            "list of examples that include all of the gold spans you plan to make "
            "predictions over."
        )

    text_ner_logger_config.user_data["nlp"] = nlp
    text_ner_logger_config.user_data["ner_config"] = nlp._pipe_configs["ner"]
    dataquality.set_labels_for_run(ner.move_names)  # type: ignore
    dataquality.set_tagging_schema(TaggingSchema.BILOU)

    nlp.add_pipe("galileo_ner")
    nlp.remove_pipe("ner")
    nlp.rename_pipe("galileo_ner", "ner")


def unwatch(nlp: Language) -> None:
    """Returns spacy nlp Language component to its original unpatched state.

    Unfortunately, spacy does not make this easy, so we replicate spacy's `add_pipe`
    for logic for using internal spacy methods to add a component object to a specific
    position.
    """
    nlp.rename_pipe("ner", "galileo_ner")
    galileo_ner = nlp.get_pipe("galileo_ner")
    if not isinstance(galileo_ner, GalileoEntityRecognizer):
        raise GalileoException(
            "Seems like your ner component has already been "
            "unwatched or was never watched in the first place."
        )

    # Spacy does not expose an easy way to add a pipeline component from an existing
    # one, so we use the internal nlp methods from `nlp.add_pipe` for it
    pipe_component = galileo_ner.__wrapped__
    pipe_component.model = galileo_ner._self_old_model
    factory_name = "ner"
    name = factory_name  # for consistency with spacy code

    pipe_index = nlp._get_pipe_index(before="galileo_ner")
    nlp._pipe_meta[name] = nlp.get_factory_meta(factory_name)
    nlp._pipe_configs[name] = text_ner_logger_config.user_data["ner_config"]
    nlp._components.insert(pipe_index, (name, pipe_component))

    nlp.remove_pipe("galileo_ner")


class GalileoEntityRecognizer(CallableObjectProxy):
    """An EntityRecognizer proxy using the wrapt library.

    We subclass the CallableObjectProxy class because we override the __call__
    method. The wrapped 'ner' object is accessible via "self.__wrapped__".
    """

    def __init__(self, ner: EntityRecognizer):
        super().__init__(ner)
        validate_obj(ner, check_type=EntityRecognizer, has_attr="model")
        if isinstance(ner, GalileoEntityRecognizer):
            raise GalileoException(
                "Seems like your ner component has already "
                "been watched. Make sure to call `watch` "
                "on a fresh `nlp` spacy Language."
            )
        # Assert we are working with the 'ner' component and not 'beam_ner'
        if self.cfg["beam_width"] != 1:
            raise GalileoException(
                f"Your EntityRecognizer's beam width is set to "
                f"{self.cfg['beam_width']}. Galileo currently "
                f"expects a beam width of 1 (the 'ner' default)."
            )

        self._self_old_model = ner.model.copy()
        ner.model = GalileoTransitionBasedParserModel(ner.model)

    def greedy_parse(self, docs: List[Doc], drop: float = 0.0) -> List:
        """Python-land implementation of the greedy_parse method in the ner component

        Transcribes the greedy_parse method in Parser to python. This allows us to call
        the Thinc model's forward function, which we patch, rather than the ridiculous
        C-math copy of it. See spacy.ml.parser_model.predict_states for the C-math fn
        that gets called eventually to do the predictions by the EntitiyRecognizer's
        greedy_parse method
        https://github.com/explosion/spaCy/blob/ed561cf428494c2b7a6790cd4b91b5326102b59d/spacy/ml/parser_model.pyx#L93
        """
        self._ensure_labels_are_added(docs)
        set_dropout_rate(self.model, drop)
        batch = self.moves.init_batch(docs)
        step_model = self.model.predict(docs)

        states = list(batch)
        non_final_states = states.copy()
        while non_final_states:
            scores = step_model.predict(non_final_states)
            self.transition_states(non_final_states, scores)  # updates non_final_states
            non_final_states = [
                state for state in non_final_states if not state.is_final()
            ]
        return states

    def predict(self, docs: Union[Doc, List[Doc]]) -> List:
        """Copied from the EntityRecognizer's predict, but calls our greedy_parse"""
        if isinstance(docs, Doc):
            docs = [docs]
        # Ignore mypy because Doc implements __len__ but Spacy doesn't realize
        # https://github.com/explosion/spaCy/blob/master/spacy/tokens/doc.pyx#L489
        if not any(filter(lambda doc: len(doc), docs)):  # type: ignore
            result = self.moves.init_batch(docs)
            return result

        # Galileo's version of greedy_parse (ner's method that handles prediction)
        return self.greedy_parse(docs, drop=0.0)

    def pipe(self, docs: Union[Doc, List[Doc]], batch_size: int = 256) -> Generator:
        """Copy of spacy EntityRecognizer's pipe

        This method copies spacy's EntityRecognizer's pipe defined in the Parser
        superclass, but notice that in calling predict from this scope it calls the
        GalileoEntityRecognizer defined above predict method.
        """
        for batch in minibatch(docs, size=batch_size):
            batch_in_order = list(batch)
            try:
                by_length = sorted(batch, key=lambda doc: len(doc))
                for subbatch in minibatch(by_length, size=max(batch_size // 4, 2)):
                    subbatch = list(subbatch)
                    parse_states = self.predict(subbatch)
                    self.set_annotations(subbatch, parse_states)
                yield from batch_in_order
            except Exception as e:
                self.get_error_handler()(self.name, self, batch_in_order, e)

    def __call__(self, doc: Doc) -> Doc:
        """Copy of TrainablePipe's __call__

        As the EntityRecognizer inherits from TrainablePipe it also inherits this
        method which calls 'predict' when the language is used as a callable.
        We need to overwrite to call our greedy_parse eventually

        e.g: nlp("some text") # calls under the hood __call__ for every pipe component
        """
        error_handler = self.get_error_handler()
        try:
            scores = self.predict([doc])
            self.set_annotations([doc], scores)
            return doc
        except Exception as e:
            return error_handler(self.name, self, [doc], e)


@Language.factory("galileo_ner")
def create_galileo_ner(nlp: Language, name: str) -> GalileoEntityRecognizer:
    return GalileoEntityRecognizer(nlp.get_pipe("ner"))


class ThincModelWrapper(CallableObjectProxy):
    """A Thinc Model obj wrapper using the wrapt library.

    wrapt primer: https://wrapt.readthedocs.io/en/latest/wrappers.html
    Wrapping an obj allows us to manipulate/watch its behavior. Compared to
    overwriting some of its functions, using an obj wrapper allows us to also
    maintain some state in between wrapped function calls. This also gives us a
    method of extending cython def classes to make them watchable. For example
    with the spacy EntityRecognizer we can wrap the class and create our own forward
    pass that calls methods we are able to log probabilities from. Successfully
    wrapping an obj relies on the ability to overwrite the ref to that obj with
    the wrapped obj, so if the obj references is immutable the better approach
    would be to overwrite the obj's behavior directly.
    wrapped object are accessible via 'self.__wrapped'

    This Thinc Model wrapper patches the 'self._func' method (the forward pass).
    """

    def __init__(self, model: thinc.model.Model):
        super().__init__(model)
        validate_obj(model, check_type=thinc.model.Model, has_attr="_func")

        self._self_orig_forward = model._func
        # https://github.com/python/mypy/issues/2427
        model._func = self._self_forward  # type: ignore

    def _self_forward(
        self, model: thinc.model.Model, X: Any, is_train: bool
    ) -> Tuple[Any, Any]:
        """Overwrite this to patch the Thinc model's forward fn"""


class GalileoTransitionBasedParserModel(ThincModelWrapper):
    expected_model_name: str = "parser_model"

    def __init__(self, model: thinc.model.Model):
        super().__init__(model)
        if not model.name == GalileoTransitionBasedParserModel.expected_model_name:
            raise GalileoException(
                "Expected the TransitionBasedParser Thinc Model to "
                f"be called {GalileoTransitionBasedParserModel.expected_model_name}. "
                f"Instead received {model.name}. This may indicate "
                f"that the spacy architecture has changed and is no "
                f"longer compatible with this Galileo integration."
            )
        assert isinstance(model, thinc.model.Model)
        assert not isinstance(model, GalileoTransitionBasedParserModel)

    def _self_forward(
        self, model: thinc.model.Model, X: Any, is_train: bool
    ) -> Tuple[ParserStepModel, Callable]:
        parser_step_model, backprop_fn = self._self_orig_forward(
            model, X, is_train=is_train
        )

        model_logger = TextNERModelLogger()
        if model_logger.logger_config.cur_split == Split.inference:
            warnings.warn(
                "Inference logging with Galileo coming soon. For now "
                "skipping logging"
            )
            return parser_step_model, backprop_fn
        if not all(["id" in doc.user_data for doc in X]):
            raise GalileoException(
                "One of your input's docs is missing a galileo generated "
                "id. Did you first log your docs/examples with us using, "
                "for example, "
                "`log_input_examples(training_examples, split='training')`? "
                "Make sure to then continue using 'training_examples'. If this "
                "is inference (nlp('some text')) then set "
                "`dataquality.set_split(split='inference', "
                "inference_name='some_name')` to continue."
            )
        helper_data = model_logger.log_helper_data
        helper_data["logits"] = {doc.user_data["id"]: [None] * len(doc) for doc in X}
        helper_data["embs"] = {doc.user_data["id"]: [None] * len(doc) for doc in X}
        helper_data["spacy_states"] = defaultdict(list)
        helper_data["spacy_states_end_idxs"] = defaultdict(list)
        helper_data["already_logged"] = False
        return GalileoParserStepModel(parser_step_model, model_logger), backprop_fn


class GalileoParserStepModel(ThincModelWrapper):
    expected_model_name: str = "parser_step_model"

    def __init__(self, model: ParserStepModel, model_logger: TextNERModelLogger):
        super().__init__(model)
        if not model.name == GalileoParserStepModel.expected_model_name:
            raise GalileoException(
                "Expected the ParserStepModel Thinc Model "
                f"to be called {GalileoParserStepModel.expected_model_name}. "
                f"Instead received {model.name}. This may indicate "
                f"that the spacy architecture has changed and is no "
                f"longer compatible with this Galileo integration."
            )
        assert isinstance(model, ParserStepModel)
        assert not isinstance(
            model, GalileoParserStepModel
        ), "trying to patch an already patched model"

        self._self_model_logger = model_logger
        # state2vec is the embedding model/layer
        model.state2vec = GalileoState2Vec(model.state2vec, self._self_model_logger)

    def _self_get_state_end_idx(self, state_idx: int, states: List[StateClass]) -> int:
        state = states[state_idx]
        state_id = state.doc.user_data["id"]
        next_state = states[state_idx + 1] if state_idx + 1 < len(states) else None
        next_state_id = next_state.doc.user_data["id"] if next_state else None

        # If the next state is from the same doc, return its start
        if next_state and next_state_id == state_id:
            return next_state.queue[0]
        # Otherwise, we're at the end of this doc, so return its length
        return len(state.doc)

    def _self_initialize_helper_data(self, states: List[StateClass]) -> None:
        # TODO: might need to assert that states has a particular structure
        helper_data = self._self_model_logger.log_helper_data
        for state_idx, state in enumerate(states):
            doc_id = state.doc.user_data["id"]
            state_end_idx = self._self_get_state_end_idx(state_idx, states)
            helper_data["spacy_states"][doc_id].append(None)
            helper_data["spacy_states_end_idxs"][doc_id].append(state_end_idx)

    def _self_is_helper_data_filled(self) -> bool:
        """Should be enough to check that logits is filled"""
        helper_data = self._self_model_logger.log_helper_data
        is_doc_filled = []
        for doc_id, doc_logits in helper_data["logits"].items():
            doc_embs = helper_data["embs"][doc_id]
            is_doc_filled.append(
                all(
                    [
                        token_logits is not None and token_embs is not None
                        for token_logits, token_embs in zip(doc_logits, doc_embs)
                    ]
                )
            )
        return all(is_doc_filled)

    def _self_fill_helper_data(
        self, states: List[StateClass], scores: np.ndarray
    ) -> None:
        helper_data = self._self_model_logger.log_helper_data
        logits = scores[..., 1:]  # Throw out the -U token
        ner = text_ner_logger_config.user_data["nlp"].get_pipe("ner")
        for state_idx, state in enumerate(states):
            doc_id = state.doc.user_data["id"]
            state_cur_token_idx = state.queue[0]

            helper_data["logits"][doc_id][state_cur_token_idx] = logits[state_idx]

            for chunk_idx, state_end_idx in enumerate(
                helper_data["spacy_states_end_idxs"][doc_id]
            ):
                if state_cur_token_idx < state_end_idx:
                    # check if this is a pre final state
                    if state_cur_token_idx == state_end_idx - 1:
                        final_state = state.copy()
                        ner.transition_states(
                            [final_state], scores[state_idx : state_idx + 1, :]
                        )
                        helper_data["spacy_states"][doc_id][chunk_idx] = final_state
                        helper_data["spacy_states_end_idxs"][doc_id][chunk_idx] += 1
                    break

    def _self_set_annotations(self, docs: Dict[int, Doc]) -> None:
        helper_data = self._self_model_logger.log_helper_data
        ner = text_ner_logger_config.user_data["nlp"].get_pipe("ner")
        for doc_id, doc in docs.items():
            states_for_doc = helper_data["spacy_states"][doc_id]
            ner.set_annotations([doc] * len(states_for_doc), states_for_doc)

    def _self_get_valid_logits(self, docs: Dict[int, Doc]) -> DefaultDict:
        helper_data = self._self_model_logger.log_helper_data
        docs_predictions = convert_spacy_ents_for_doc_to_predictions(
            docs, self._self_model_logger.logger_config.labels
        )
        docs_valid_logits = defaultdict(list)
        for doc_id, doc_logits in helper_data["logits"].items():
            for token_idx, token_logits in enumerate(doc_logits):
                valid_token_logits = convert_spacy_ner_logits_to_valid_logits(
                    token_logits, docs_predictions[doc_id][token_idx]
                )
                docs_valid_logits[doc_id].append(valid_token_logits)
        return docs_valid_logits

    def _self_populate_model_logger(self, docs_valid_logits: DefaultDict) -> None:
        model_logger = self._self_model_logger
        helper_data = model_logger.log_helper_data
        model_logger.ids = list(model_logger.ids)  # for mypy's sake
        for doc_id, doc_valid_logits in docs_valid_logits.items():
            model_logger.logits.append(np.array(doc_valid_logits))
            doc_embs = helper_data["embs"][doc_id]
            model_logger.embs.append(np.array(doc_embs))
            model_logger.ids.append(doc_id)

    def _self_get_docs_copy(self) -> Dict[int, Doc]:
        spacy_states = self._self_model_logger.log_helper_data["spacy_states"]
        return {
            doc_id: states_for_doc[0].doc.copy()
            for doc_id, states_for_doc in spacy_states.items()
        }

    def _self_forward(
        self,
        parser_step_model: ParserStepModel,
        states: List[StateClass],
        is_train: bool,
    ) -> Tuple[np.ndarray, Callable]:
        """Patches the ParserStepModel's forward func

        The forward function is called with a token from each chunked doc in the
        sample. So the max num times its called is equal to the max num tokens in any
        doc's chunk.
        """
        helper_data = self._self_model_logger.log_helper_data
        if not helper_data["spacy_states"]:
            self._self_initialize_helper_data(states)

        parser_step_model.state2vec._self_states = states
        # Let spacy run its original forward on the data
        scores, backprop_fn = self._self_orig_forward(
            parser_step_model, states, is_train
        )

        self._self_fill_helper_data(states, scores.copy())

        if not helper_data["already_logged"] and self._self_is_helper_data_filled():
            docs_copy = self._self_get_docs_copy()
            self._self_set_annotations(docs_copy)
            docs_valid_logits = self._self_get_valid_logits(docs_copy)

            # Now that we have valid logits and embs, fill out the log
            self._self_populate_model_logger(docs_valid_logits)
            self._self_model_logger.log()
            helper_data["already_logged"] = True

        return scores, backprop_fn


class GalileoState2Vec(CallableObjectProxy):
    def __init__(self, model: State2Vec, model_logger: TextNERModelLogger):
        super().__init__(model)
        validate_obj(model, State2Vec, "__call__")
        assert not isinstance(
            model, GalileoState2Vec
        ), "trying to patch an already patched model"
        self._self_model_logger = model_logger
        self._self_states: List[StateClass] = []  # Set externally by developer

    def __call__(self, *args: Any, **kwargs: Any) -> Tuple[np.ndarray, np.ndarray]:
        """Overwrites forward to capture embeddings and add to model_logger.

        Note that self._self_states needs to be set by developer before this is called.
        """
        embeddings, embeddings_bp = self.__wrapped__(*args, **kwargs)

        helper_data = self._self_model_logger.log_helper_data
        for state_idx, state in enumerate(self._self_states):
            doc_id = state.doc.user_data["id"]
            state_cur_token_idx = state.queue[0]
            helper_data["embs"][doc_id][state_cur_token_idx] = embeddings[
                state_idx
            ].copy()

        return embeddings, embeddings_bp



import warnings
from abc import abstractmethod
from typing import Dict, List, Optional, Union
from uuid import uuid4

import numpy as np
from scipy.special import expit, softmax

from dataquality import config
from dataquality.exceptions import GalileoException
from dataquality.loggers.base_logger import BaseGalileoLogger
from dataquality.loggers.data_logger import BaseGalileoDataLogger
from dataquality.schemas.split import Split
from dataquality.schemas.task_type import TaskType
from dataquality.utils.dq_logger import get_dq_logger
from dataquality.utils.thread_pool import ThreadPoolManager
from dataquality.utils.vaex import _save_hdf5_file


class BaseGalileoModelLogger(BaseGalileoLogger):
    def __init__(
        self,
        embs: Union[List, np.ndarray] = None,
        probs: Union[List, np.ndarray] = None,
        logits: Union[List, np.ndarray] = None,
        ids: Union[List, np.ndarray] = None,
        split: str = "",
        epoch: Optional[int] = None,
        inference_name: Optional[str] = None,
    ) -> None:
        super().__init__()
        # Need to compare to None because they may be np arrays which cannot be
        # evaluated with bool directly
        self.embs: Union[List, np.ndarray] = embs if embs is not None else []
        self.logits: Union[List, np.ndarray] = logits if logits is not None else []
        self.probs: Union[List, np.ndarray] = probs if probs is not None else []
        self.ids: Union[List, np.ndarray] = ids if ids is not None else []
        self.epoch = epoch
        self.split: str = split
        self.inference_name = inference_name

    def _log(self) -> None:
        """Threaded logger target"""
        try:
            self.validate()
        except AssertionError as e:
            get_dq_logger().error(
                "Validation of data failed", split=self.split, epoch=self.epoch
            )
            raise GalileoException(
                f"The provided logged data is invalid: {e}"
            ) from None
        data = self._get_data_dict()
        self.write_model_output(data)

    def _add_threaded_log(self) -> None:
        try:
            self._log()
        except Exception as e:
            get_dq_logger().exception(
                "Logging of model outputs failed", split=self.split, epoch=self.epoch
            )
            err_msg = (
                f"An issue occurred while logging model outputs. Address any issues in "
                f"your logging and make sure to call dq.init before restarting:\n"
                f"{str(e)}"
            )
            warnings.warn(err_msg)
            self.logger_config.exception = err_msg

    def log(self) -> None:
        """The top level log function that try/excepts its child"""
        self.check_for_logging_failures()
        # We validate split and epoch before entering the thread because we reference
        # global variables (cur_split and cur_epoch) that are subject to change
        # between subsequent threads
        self.set_split_epoch()
        get_dq_logger().info(
            "Starting logging process from thread", split=self.split, epoch=self.epoch
        )
        ThreadPoolManager.add_thread(target=self._add_threaded_log)

    def write_model_output(self, data: Dict) -> None:
        """Creates an hdf5 file from the data dict"""
        get_dq_logger().info("Writing model output", split=self.split, epoch=self.epoch)
        location = (
            f"{self.LOG_FILE_DIR}/{config.current_project_id}"
            f"/{config.current_run_id}"
        )
        epoch, split = data["epoch"][0], data["split"][0]

        if split == Split.inference:
            inference_name = data["inference_name"][0]
            path = f"{location}/{split}/{inference_name}"
        else:
            path = f"{location}/{split}/{epoch}"

        object_name = f"{str(uuid4()).replace('-', '')[:12]}.hdf5"
        get_dq_logger().info("Saving hdf5 file", split=self.split)
        _save_hdf5_file(path, object_name, data)

    def set_split_epoch(self) -> None:
        super().set_split_epoch()
        if self.split == Split.inference and self.inference_name is None:
            if self.logger_config.cur_inference_name is not None:
                self.inference_name = self.logger_config.cur_inference_name
            else:
                raise GalileoException(
                    "For inference split you must either log an inference name "
                    "or set it before logging. Use `dataquality.set_split` to set"
                    "inference_name"
                )
        # Epoch can be ignored for inference split
        if self.split != Split.inference and self.epoch is None:
            if self.logger_config.cur_epoch is not None:
                self.epoch = self.logger_config.cur_epoch
            else:
                raise GalileoException(
                    "You must either log an epoch or set it before logging. Use "
                    "`dataquality.set_epoch` to set the epoch"
                )

    @classmethod
    def upload(cls) -> None:
        """The upload function is implemented in the sister DataConfig class"""
        BaseGalileoDataLogger.get_logger(TaskType[cls.__logger_name__]).upload()

    @staticmethod
    def get_model_logger_attr(cls: object) -> str:
        """
        Returns the attribute that corresponds to the GalileoModelLogger class.
        This assumes only 1 GalileoModelLogger object exists in the class

        :param cls: The class
        :return: The attribute name
        """
        for attr in dir(cls):
            member_class = getattr(cls, attr)
            if isinstance(member_class, BaseGalileoModelLogger):
                return attr
        raise AttributeError("No model logger attribute found!")

    @abstractmethod
    def _get_data_dict(self) -> Dict:
        """Constructs a dictionary of arrays from logged model output data"""

    def convert_logits_to_prob_binary(self, sample_logits: np.ndarray) -> np.ndarray:
        """Converts logits to probs in the binary case

        Takes the sigmoid of the single class logits and adds the negative
        lass prediction (1-class pred)
        """
        sample_probs = expit(sample_logits)
        probs_1 = np.expand_dims(sample_probs, axis=-1)
        probs_0 = 1 - probs_1
        probs = np.concatenate([probs_0, probs_1], axis=-1)
        return probs

    def convert_logits_to_probs(
        self, sample_logits: Union[List[np.ndarray], np.ndarray]
    ) -> np.ndarray:
        """Converts logits to probs via softmax"""
        # axis ensures that in a matrix of probs with dims num_samples x num_classes
        # we take the softmax for each sample
        if not isinstance(sample_logits, np.ndarray):
            sample_logits = self._convert_tensor_ndarray(sample_logits)
        if len(sample_logits.shape) == 1 or sample_logits.shape[1] == 1:
            if len(sample_logits.shape) > 1:
                # Remove final empty dimension if it's there
                sample_logits = sample_logits.reshape(-1)
            return self.convert_logits_to_prob_binary(sample_logits)
        return softmax(np.array(sample_logits), axis=-1)



from typing import Any, Dict, List, Tuple

import numpy as np
from pydantic import validator

from dataquality.loggers.logger_config.base_logger_config import BaseLoggerConfig


class TextNERLoggerConfig(BaseLoggerConfig):
    gold_spans: Dict[str, List[Tuple[int, int, str]]] = {}
    sample_length: Dict[str, int] = {}
    user_data: Dict[str, Any] = {}

    class Config:
        validate_assignment = True
        arbitrary_types_allowed = True

    def get_sample_key(self, split: str, sample_id: int) -> str:
        return f"{split}_{sample_id}"

    @validator("labels", always=True, pre=True, allow_reuse=True)
    def clean_labels(cls, labels: List[str]) -> List[str]:
        if isinstance(labels, np.ndarray):
            labels = labels.tolist()
        if labels is not None:
            assert isinstance(labels, List), "Labels must be a list"
        if labels and len(labels) == 1:
            labels = [labels[0], f"NOT_{labels[0]}"]
        return labels


text_ner_logger_config = TextNERLoggerConfig()



from enum import Enum, unique

from pydantic import BaseModel
from vaex.dataframe import DataFrame


class BaseLoggerInOutFrames(BaseModel):
    prob: DataFrame
    emb: DataFrame
    data: DataFrame

    class Config:
        arbitrary_types_allowed = True


@unique
class FileType(str, Enum):
    """Valid file extensions for an exported dataframe"""

    arrow = "arrow"
    parquet = "parquet"
    json = "json"
    csv = "csv"



import os
import warnings
from abc import abstractmethod
from typing import Any, Dict, Iterable, List, Optional, Tuple, TypeVar, Union
from uuid import uuid4

import numpy as np
import pandas as pd
import pyarrow as pa
import vaex
from vaex.dataframe import DataFrame

from dataquality.clients.objectstore import ObjectStore
from dataquality.core._config import config
from dataquality.exceptions import GalileoException, GalileoWarning
from dataquality.loggers.base_logger import BaseGalileoLogger, BaseLoggerAttributes
from dataquality.schemas.dataframe import BaseLoggerInOutFrames
from dataquality.schemas.ner import TaggingSchema
from dataquality.schemas.split import Split
from dataquality.utils import tqdm
from dataquality.utils.hdf5_store import HDF5_STORE
from dataquality.utils.thread_pool import ThreadPoolManager
from dataquality.utils.vaex import (
    _join_in_out_frames,
    concat_hdf5_files,
    filter_df,
    validate_ids_for_df,
    validate_unique_ids,
)

DATA_FOLDERS = ["emb", "prob", "data"]
DataSet = TypeVar("DataSet", bound=Union[Iterable, pd.DataFrame, DataFrame])
MetasType = TypeVar("MetasType", bound=Dict[str, List[Union[str, float, int]]])
MetaType = TypeVar("MetaType", bound=Dict[str, Union[str, float, int]])
ITER_CHUNK_SIZE = 100_000


class BaseGalileoDataLogger(BaseGalileoLogger):
    MAX_META_COLS = 50  # Limit the number of metadata attrs a user can log
    MAX_STR_LEN = 50  # Max characters in a string metadata attribute
    INPUT_DATA_NAME = "input_data.arrow"

    DATA_FOLDER_EXTENSION = {data_folder: "hdf5" for data_folder in DATA_FOLDERS}

    def __init__(self, meta: MetasType = None) -> None:
        super().__init__()
        self.meta: Dict = meta or {}
        self.log_export_progress = True

    @abstractmethod
    def log_data_sample(self, *, text: str, id: int, **kwargs: Any) -> None:
        """Log a single input sample. See child for details"""

    @abstractmethod
    def log_data_samples(
        self, *, texts: List[str], ids: List[int], **kwargs: Any
    ) -> None:
        """Log a list of input samples. See child for details"""

    @abstractmethod
    def log_dataset(
        self,
        dataset: DataSet,
        *,
        batch_size: int = ITER_CHUNK_SIZE,
        text: Union[str, int] = "text",
        id: Union[str, int] = "id",
        split: Optional[Split] = None,
        meta: Optional[List[Union[str, int]]] = None,
        **kwargs: Any,
    ) -> None:
        """Log a dataset/iterable of input samples.

        Provide the dataset and the keys to index into it. See child for details"""

    def log(self) -> None:
        """Writes input data to disk in .galileo/logs

        If input data already exist, append new data to existing input file
        """
        self.validate()
        write_input_dir = (
            f"{BaseGalileoLogger.LOG_FILE_DIR}/{config.current_project_id}/"
            f"{config.current_run_id}"
        )
        if not os.path.exists(write_input_dir):
            os.makedirs(write_input_dir)

        df = self._get_input_df()
        file_path = f"{write_input_dir}/{BaseGalileoDataLogger.INPUT_DATA_NAME}"

        if os.path.isfile(file_path):
            self.append_input_data(df, write_input_dir, file_path)
        else:
            if self.log_export_progress:
                with vaex.progress.tree("vaex", title="Exporting input data"):
                    df.export(file_path)
            else:
                df.export(file_path)
        df.close()

    def append_input_data(
        self, df: DataFrame, write_input_dir: str, file_path: str
    ) -> None:
        # Create a temporary file for existing input data
        tmp_name = f"{write_input_dir}/{str(uuid4()).replace('-', '')[:12]}.arrow"
        os.rename(file_path, tmp_name)

        # Merge existing data with new data
        existing_df = vaex.open(tmp_name)
        merged_df = vaex.concat([existing_df, df])

        try:  # Validate there are no duplicated IDs
            validate_ids_for_df(merged_df)
        except GalileoException as e:  # Cleanup and raise on error
            merged_df.close()
            os.rename(tmp_name, file_path)  # Revert name, we aren't logging
            raise e

        if self.log_export_progress:
            with vaex.progress.tree("vaex", title="Appending input data"):
                merged_df.export(file_path)
        else:
            merged_df.export(file_path)

        # Cleanup temporary file after appending input data
        os.remove(tmp_name)

    @classmethod
    def upload(cls, last_epoch: Optional[int] = None) -> None:
        """
        Iterates through all of each splits children folders [data/emb/prob] for each
        inference name / epoch, concatenates all of the files with vaex, and uploads
        them to a single file in minio
        """
        ThreadPoolManager.wait_for_threads()
        cls.check_for_logging_failures()
        print("☁️ Uploading Data")
        object_store = ObjectStore()
        proj_run = f"{config.current_project_id}/{config.current_run_id}"
        location = f"{cls.LOG_FILE_DIR}/{proj_run}"

        in_frame = vaex.open(
            f"{location}/{BaseGalileoDataLogger.INPUT_DATA_NAME}"
        ).copy()
        for split in Split.get_valid_attributes():
            split_loc = f"{location}/{split}"
            if not os.path.exists(split_loc):
                continue
            if not len(in_frame[in_frame["split"] == split]):
                warnings.warn(
                    f"There was output data logged for split {split} but no input data "
                    "logged. Skipping upload for this split as there are no samples "
                    "to join to.",
                    GalileoWarning,
                )
                continue

            in_frame_split = filter_df(in_frame, "split", split)
            cls.upload_split(object_store, in_frame_split, split, split_loc, last_epoch)

    @classmethod
    def upload_split(
        cls,
        object_store: ObjectStore,
        in_frame: DataFrame,
        split: str,
        split_loc: str,
        last_epoch: Optional[int] = None,
    ) -> None:
        # If set, last_epoch will only let you upload to and including the provided
        # epoch value, nothing more.
        # If None, then slicing a list [:None] will include all values
        epochs_or_infs = os.listdir(split_loc)
        epochs_or_infs = sorted(
            epochs_or_infs, key=lambda i: int(i) if split != Split.inference else i
        )
        # last_epoch is inclusive
        last_epoch = last_epoch + 1 if last_epoch else last_epoch
        epochs_or_infs = epochs_or_infs[:last_epoch]

        # For each inference name or epoch of the given split
        for split_run in tqdm(epochs_or_infs, total=len(epochs_or_infs), desc=split):
            in_frame_slice = in_frame.copy()
            prob_only = cls.prob_only(epochs_or_infs, split, split_run)
            if split == Split.inference:
                in_frame_slice = filter_df(in_frame_slice, "inference_name", split_run)

            dir_name = f"{split_loc}/{split_run}"
            in_out_frames = cls.create_in_out_frames(
                in_frame_slice, dir_name, prob_only, split, split_run
            )
            cls.upload_in_out_frames(object_store, in_out_frames, split, split_run)

    @classmethod
    def create_in_out_frames(
        cls,
        in_frame: DataFrame,
        dir_name: str,
        prob_only: bool,
        split: str,
        split_run: Union[str, int],
    ) -> BaseLoggerInOutFrames:
        str_cols = concat_hdf5_files(dir_name, prob_only)
        out_frame = vaex.open(f"{dir_name}/{HDF5_STORE}")

        if split == Split.inference:
            dtype: Union[str, None] = "str"
            epoch_or_inf_name = "inference_name"
        else:
            dtype = None
            epoch_or_inf_name = "epoch"

        # Post concat, string columns come back as bytes and need conversion
        for col in str_cols:
            out_frame[col] = out_frame[col].to_arrow().cast(pa.large_string())
        if prob_only:
            out_frame["split"] = vaex.vconstant(
                split, length=len(out_frame), dtype="str"
            )
            out_frame[epoch_or_inf_name] = vaex.vconstant(
                split_run, length=len(out_frame), dtype=dtype
            )

        return cls.process_in_out_frames(
            in_frame, out_frame, prob_only, epoch_or_inf_name
        )

    @classmethod
    def process_in_out_frames(
        cls,
        in_frame: DataFrame,
        out_frame: DataFrame,
        prob_only: bool,
        epoch_or_inf_name: str,
    ) -> BaseLoggerInOutFrames:
        """Processes input and output dataframes from logging

        Validates uniqueness of IDs in the dataframes
        Joins inputs and outputs
        Splits the dataframes into prob, emb, and data for uploading to minio
        """
        validate_unique_ids(out_frame, epoch_or_inf_name)
        in_out = _join_in_out_frames(in_frame, out_frame)

        prob, emb, data_df = cls.split_dataframe(in_out, prob_only)
        # These df vars will be used in upload_in_out_frames
        emb.set_variable("skip_upload", prob_only)
        data_df.set_variable("skip_upload", prob_only)
        epoch_inf_val = out_frame[[epoch_or_inf_name]][0][0]
        prob.set_variable("progress_name", str(epoch_inf_val))

        return BaseLoggerInOutFrames(prob=prob, emb=emb, data=data_df)

    @classmethod
    def upload_in_out_frames(
        cls,
        object_store: ObjectStore,
        in_out_frames: BaseLoggerInOutFrames,
        split: str,
        split_run: Union[str, int],
    ) -> None:
        proj_run = f"{config.current_project_id}/{config.current_run_id}"

        prob = in_out_frames.prob
        emb = in_out_frames.emb
        data_df = in_out_frames.data

        epoch_inf = prob.variables.pop("progress_name", "")

        name = "inf_name" if split == Split.inference else "epoch"
        desc = f"{split} ({name}={epoch_inf})"
        for data_folder, df_obj in tqdm(
            zip(DATA_FOLDERS, [emb, prob, data_df]), total=3, desc=desc, leave=False
        ):
            if df_obj.variables.get("skip_upload"):
                continue

            ext = cls.DATA_FOLDER_EXTENSION[data_folder]
            minio_file = (
                f"{proj_run}/{split}/{split_run}/{data_folder}/{data_folder}.{ext}"
            )
            object_store.create_project_run_object_from_df(
                df=df_obj, object_name=minio_file
            )

    @classmethod
    def prob_only(
        cls, epochs: List[str], split: str, split_run: Union[int, str]
    ) -> bool:
        if split == Split.inference:
            return False

        # If split is not inference, split_run must be epoch
        epoch = int(split_run)
        # For all epochs that aren't the last 2 (early stopping), we only
        # want to upload the probabilities (for DEP calculation).
        max_epoch_for_split = max([int(i) for i in epochs])
        return bool(epoch < max_epoch_for_split - 1)

    def validate(self) -> None:
        self.set_split_epoch()

    @classmethod
    @abstractmethod
    def validate_labels(cls) -> None:
        ...

    def validate_metadata(self, batch_size: int) -> None:
        if len(self.meta.keys()) > self.MAX_META_COLS:
            warnings.warn(
                f"You can only log up to {self.MAX_META_COLS} metadata attrs. "
                f"The first {self.MAX_META_COLS} will be logged only.",
                GalileoWarning,
            )
        # When logging metadata columns, if the user breaks a rule, don't fail
        # completely, just warn them and remove that metadata column
        # Cast to list for in-place dictionary mutation
        reserved_keys = BaseLoggerAttributes.get_valid()
        valid_meta = {}
        for key, values in list(self.meta.items())[: self.MAX_META_COLS]:
            # Key must not override a default
            if key in reserved_keys:
                warnings.warn(
                    f"Metadata column names must not override default values "
                    f"{reserved_keys}. Metadata field {key} "
                    f"will be removed.",
                    GalileoWarning,
                )
                continue
            bad_prefixes = ["galileo", "prob", "gold", "pred"]
            for bad_start in bad_prefixes:
                if key.startswith(bad_start):
                    warnings.warn(
                        "Metadata name must not start with the following "
                        f"prefixes: (galileo_, prob_, gold_. Won't log {key}",
                        GalileoWarning,
                    )
                    continue
            # Must be the same length as input
            if len(values) != batch_size:
                warnings.warn(
                    f"Expected {batch_size} values for key {key} but got "
                    f"{len(values)}. Will not log this metadata column.",
                    GalileoWarning,
                )
                continue
            # Values must be a point, not an iterable
            valid_types = (str, int, float, np.floating, np.integer)
            invalid_values = filter(
                lambda t: not isinstance(t, valid_types)
                or (isinstance(t, str) and len(t) > self.MAX_STR_LEN),
                values,
            )
            bad_val = next(invalid_values, None)
            if bad_val:
                warnings.warn(
                    f"Metadata column {key} has one or more invalid values {bad_val} "
                    f"of type {type(bad_val)}. Only strings of "
                    f"len < {self.MAX_STR_LEN} and numbers can be logged.",
                    GalileoWarning,
                )
                continue
            valid_meta[key] = values
        self.meta = valid_meta

    @staticmethod
    def get_data_logger_attr(cls: object) -> str:
        """
        Returns the attribute that corresponds to the logger in the class.
        This assumes only 1 logger object exists in the class

        :param cls: The class
        :return: The attribute name
        """
        for attr in dir(cls):
            member_class = getattr(cls, attr)
            if isinstance(member_class, BaseGalileoDataLogger):
                return attr
        raise AttributeError("No data logger attribute found!")

    @classmethod
    @abstractmethod
    def split_dataframe(
        cls, df: DataFrame, prob_only: bool
    ) -> Tuple[DataFrame, DataFrame, DataFrame]:
        ...

    def validate_kwargs(self, kwargs: Dict) -> None:
        """Raises if a function that shouldn't get kwargs gets any"""
        if kwargs.keys():
            raise GalileoException(f"Unexpected arguments: {tuple(kwargs.keys())}")

    @abstractmethod
    def _get_input_df(self) -> DataFrame:
        ...

    @classmethod
    def set_tagging_schema(cls, tagging_schema: TaggingSchema) -> None:
        """Sets the tagging schema, if applicable. Must be implemented by child"""
        raise GalileoException(f"Cannot set tagging schema for {cls.__logger_name__}")



from typing import Any, Dict, Optional

from pydantic import BaseModel, validator

from dataquality.schemas.ner import TaggingSchema
from dataquality.schemas.split import Split


class BaseLoggerConfig(BaseModel):
    labels: Any = None
    tasks: Any = None
    observed_num_labels: Any = None
    observed_labels: Any = None
    tagging_schema: Optional[TaggingSchema]
    last_epoch: int = 0
    cur_epoch: Optional[int]
    cur_split: Optional[Split]
    cur_inference_name: Optional[str]
    training_logged: bool = False
    validation_logged: bool = False
    test_logged: bool = False
    inference_logged: bool = False
    exception: str = ""

    class Config:
        validate_assignment = True

    def reset(self) -> None:
        """Reset all class vars"""
        self.__init__()  # type: ignore

    @validator("cur_split")
    def inference_sets_inference_name(
        cls, field_value: Split, values: Dict[str, Any]
    ) -> Split:
        if field_value == Split.inference:
            assert values.get(
                "cur_inference_name"
            ), "Please specify inference_name when setting split to inference"
        return field_value


base_logger_config = BaseLoggerConfig()



import os
import shutil
from typing import Any, Callable, Dict, Generator, List
from unittest.mock import MagicMock
from uuid import UUID

import pytest
import requests
import spacy
from vaex.dataframe import DataFrame

import dataquality
from dataquality import config
from dataquality.clients import objectstore
from dataquality.loggers import BaseGalileoLogger
from dataquality.schemas.task_type import TaskType
from dataquality.utils.dq_logger import DQ_LOG_FILE_HOME

DEFAULT_API_URL = "http://localhost:8088"
DEFAULT_PROJECT_ID = UUID("399057bc-b276-4027-a5cf-48893ac45388")
DEFAULT_RUN_ID = UUID("399057bc-b276-4027-a5cf-48893ac45388")

LOCATION = f"{BaseGalileoLogger.LOG_FILE_DIR}/{DEFAULT_PROJECT_ID}/{DEFAULT_RUN_ID}"
DQ_LOG_FILE_LOCATION = f"{DQ_LOG_FILE_HOME}/{DEFAULT_RUN_ID}"
TEST_STORE_DIR = "TEST_STORE"
TEST_PATH = f"{LOCATION}/{TEST_STORE_DIR}"
SPLITS = ["training", "test"]
SUBDIRS = ["data", "emb", "prob"]

spacy.util.fix_random_seed()


@pytest.fixture(autouse=True)
def disable_network_calls(request, monkeypatch):
    # Tests that fetch datasets need network access
    if "noautofixt" in request.keywords:
        return

    def stunted_get(url: str) -> MagicMock:
        """Unless it's a mocked call to healthcheck, disable network access"""
        if url.endswith("healthcheck"):
            mock = MagicMock(ok=True)
            return mock
        raise RuntimeError("Network access not allowed during testing!")

    monkeypatch.setattr(requests, "get", lambda url, *args, **kwargs: stunted_get(url))


@pytest.fixture(scope="function")
def cleanup_after_use() -> Generator:
    try:
        dataquality.get_model_logger().logger_config.reset()
        if os.path.isdir(BaseGalileoLogger.LOG_FILE_DIR):
            shutil.rmtree(BaseGalileoLogger.LOG_FILE_DIR)
        if not os.path.isdir(TEST_PATH):
            for split in SPLITS:
                for subdir in SUBDIRS:
                    os.makedirs(f"{TEST_PATH}/{split}/{subdir}")
        if not os.path.isdir(DQ_LOG_FILE_LOCATION):
            os.makedirs(DQ_LOG_FILE_LOCATION)
        yield
    finally:
        shutil.rmtree(BaseGalileoLogger.LOG_FILE_DIR)
        shutil.rmtree(DQ_LOG_FILE_LOCATION)
        dataquality.get_model_logger().logger_config.reset()


@pytest.fixture()
def set_test_config(
    default_token: str = "sometoken",
    default_task_type: TaskType = TaskType.text_classification,
    default_api_url: str = DEFAULT_API_URL,
) -> Callable:
    config.token = default_token
    config.task_type = default_task_type
    config.api_url = default_api_url
    config.current_run_id = DEFAULT_RUN_ID
    config.current_project_id = DEFAULT_PROJECT_ID

    def curry(**kwargs: Dict[str, Any]) -> None:
        # Override test config with custom value by currying
        for k, v in kwargs.items():
            if k in config.dict():
                config.__setattr__(k, v)
        dataquality.get_model_logger().logger_config.reset()

    return curry


@pytest.fixture()
def statuses_response() -> Dict[str, List]:
    return {
        "statuses": [
            {"status": "started", "timestamp": "2022-02-20"},
            {"status": "finished", "timestamp": "2022-02-24"},
        ]
    }


@pytest.fixture()
def input_data() -> Callable:
    def curry(
        split: str = "training",
        inference_name: str = "all-customers",
        meta: Dict = None,
    ) -> Dict:
        data = {
            "texts": ["sentence_1", "sentence_2"],
            "split": split,
            "ids": [1, 2],
        }
        if split == "inference":
            data.update(inference_name=inference_name)
        else:
            data.update(labels=["APPLE", "ORANGE"])

        if meta:
            data.update(meta=meta)

        return data

    return curry


def patch_object_upload(self: Any, df: DataFrame, object_name: str) -> None:
    """
    A patch for the object_store.create_project_run_object_from_df so we don't have to
    talk to minio for testing
    """
    # separate folder per split (test, train, val) and data type (emb, prob, data)
    split, epoch, data_type, file_name = object_name.split("/")[-4:]
    export_path = f"{TEST_PATH}/{split}/{epoch}/{data_type}"
    export_loc = f"{export_path}/{file_name}"

    if not os.path.isdir(export_path):
        os.makedirs(export_path)
    df.export(export_loc)


# Patch the upload so we don't write to S3/minio
objectstore.ObjectStore.create_project_run_object_from_df = patch_object_upload



from enum import Enum
from typing import Optional, Union

from pydantic import UUID4

from dataquality.schemas.split import Split


class Route(str, Enum):
    """
    List of available API routes
    """

    projects = "projects"
    runs = "runs"
    users = "users"
    cleanup = "cleanup"
    login = "login"
    current_user = "current_user"
    healthcheck = "healthcheck/runners"
    slices = "slices"
    split_path = "split"
    jobs = "jobs"
    presigned_url = "presigned_url"
    tasks = "tasks"
    labels = "labels"
    epochs = "epochs"
    summary = "insights/summary"
    groupby = "insights/groupby"
    distribution = "insights/distribution"
    xray = "insights/xray"
    export = "export"
    edits = "edits"
    export_edits = "edits/export"

    @staticmethod
    def content_path(
        project_id: Optional[Union[str, UUID4]] = None,
        run_id: Optional[Union[str, UUID4]] = None,
        split: Optional[Union[str, Split]] = None,
    ) -> str:
        path = ""
        if project_id:
            path += f"/{Route.projects}/{project_id}"
        if run_id:
            path += f"/{Route.runs}/{run_id}"
        if split:
            path += f"/{Route.split_path}/{split}"

        return path.strip("/")



from typing import Dict, Optional


def headers(token: Optional[str]) -> Dict[str, str]:
    if not token:
        raise ValueError(
            "Missing token passed to headers utility! "
            "Check that you are logged into Galileo by running "
            "dataquality.config.email. "
            "Your account email should appear."
        )
    return {"Authorization": f"Bearer {token}"}



import os
from functools import wraps
from typing import Callable, Optional, TypeVar

from typing_extensions import ParamSpec

T = TypeVar("T")
P = ParamSpec("P")
GALILEO_DISABLED = "GALILEO_DISABLED"


def check_noop(func: Callable[P, T]) -> Callable[P, Optional[T]]:
    """Checks if GALILEO_DISABLED is set. If so, skip the function call

    https://peps.python.org/pep-0612/
    """

    # Wrap is used to preserve the docstring of the original function
    @wraps(func)
    def decorator(*args: P.args, **kwargs: P.kwargs) -> Optional[T]:
        if galileo_disabled():
            return None
        return func(*args, **kwargs)

    return decorator


def galileo_disabled() -> bool:
    return os.getenv(GALILEO_DISABLED) in (True, "TRUE", "True", "true", 1)
